#!/usr/bin/env python3
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2023  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##

import os
import logging
from passlib.hash import sha512_crypt, bcrypt
from passlib.hash import pbkdf2_sha256
from passlib.apache import HtpasswdFile
import datetime
import base64
import binascii
import ldap3.utils

from waptutils import datetime2isodate

from waptserver.common import get_secured_token_generator, get_server_uuid
from waptserver.utils import EWaptAuthenticationFailure

from waptserver.app import app

from waptserver.config import DEFAULT_CONFIG_DIR, rewrite_config_item
from waptserver.model import wapt_db, WaptUserAcls
from itsdangerous import BadData

import auth_module_ad
import time

logger = logging.getLogger('waptauth')


def valid_username(username):
    return username is not None and username != '' and len(username) < 255

def get_htpasswdfile() -> HtpasswdFile:
    htpasswd_path = ''
    if app.conf.get('htpasswd_path'):
        htpasswd_path = app.conf.get('htpasswd_path')
    else:
        root = app.conf.get('application_root')

        if root == '':
            htpasswd_path = os.path.join(DEFAULT_CONFIG_DIR, 'htpasswd')
        else:
            htpasswd_path = os.path.join(DEFAULT_CONFIG_DIR, root + '_htpasswd')
    return HtpasswdFile(htpasswd_path, new = not os.path.isfile(htpasswd_path))

def set_user_password_htpasswd(username,new_password,old_password=None):
    """

    Returns:
        bool: True if user updated, added, False if user added

    """
    if username.lower() == app.conf['wapt_user'].lower():
        raise EWaptAuthenticationFailure('admin password can not be changed here')

    if not username or not new_password:
        raise EWaptAuthenticationFailure('no username or password provided')

    htpasswd_users = get_htpasswdfile()
    if old_password is None or htpasswd_users.check_password(username, old_password):
        result = htpasswd_users.set_password(username,new_password)
        htpasswd_users.save()
        return result
    else:
        raise EWaptAuthenticationFailure('old password failure')

def delete_user_password_htpasswd(username,old_password=None):
    """

    Returns:
        bool: True if user updated, added, False if user added

    """
    if username.lower() == app.conf['wapt_user'].lower():
        raise EWaptAuthenticationFailure('admin password can not be changed here')

    if not username:
        raise EWaptAuthenticationFailure('no username provided')

    htpasswd_users = get_htpasswdfile()
    if old_password is None or htpasswd_users.check_password(username, old_password):
        result = htpasswd_users.delete(username)
        htpasswd_users.save()
        return result
    else:
        raise EWaptAuthenticationFailure('old password failure')

def list_user_password_htpasswd():
    return get_htpasswdfile().users()

def get_cn_from_x_ssl_client_auth(request):
    if request.headers.get('X-Ssl-Authenticated', None) == 'SUCCESS':
        dn = request.headers.get('X-Ssl-Client-Dn', None)
        if dn:
            dn_parts = ldap3.utils.dn.parse_dn(dn)
            for code,value,sep in dn_parts:
                if code=='CN':
                    return value
    return None

def check_auth(username=None, password=None, request=None,
               session=None, methods=None):
    """This function is called to check if a username /
    password combination is valid or to get already authenticated username
    from session argument.

    If no username or password, use session.authorization header if available.

    Args:
        username (str or None): user requiring to login.
                        If None, and request arg is not None, username is taken from request.authorization
        password (str or None):
                        If None, and request arg is not None, password is taken from request.authorization
        request (Flask.Request or None) : request where to lookup authorization basic header
        session (Flask.Session or None) where to lookup session user
        methods (list of str) : list of auth method to try until one is successfull
                               'session': use current session user if it matches username
                               'admin'  : use 'wapt_password' parameter stored in waptserver.ini if username matches the 'wapt_user' config parameter.
                               'ldap'   : makes a ldap authenticated bind to 'ldap_auth_server' with username and password arguments.
                               'kerb'   :
                               'ssl'    :
                               'passwd' : username and passwords are checked against the htpasswd file located at 'htpasswd_path' config parameter.
                                          username can NOT be equal to 'wapt_user' config parameter.
                               'token'  : token is extracted from Authorization Bearer header. if No Authorization Bearer, token is equal to password
                                          username must be equal to the 'server_uuid' or to the 'user' key of token content.
                                          'server_uuid' token key must be equal to the current server_uuid.
                                          token validity is checked against 'token_secret' and 'token_lifetime' parameters.

    Returns:
        dict (auth_method = auth_method,user=auth_user,auth_date=auth_date)
             None if auth is not successfull

    """
    if methods is None:
        methods = app.conf['default_auth_methods']

    if not password:
        password = None

    if username is None and password is None and request is not None and request.authorization:
        username = request.authorization.username
        password = request.authorization.password

    if password and password[0:4] == 'b64_':
        try:
            password = base64.decodestring(str.encode(password[4:]))
        except binascii.Error:
            pass

    auth_method = None
    auth_user = None
    auth_date = None

    assert(isinstance(methods, list))

    for method in methods:
        logger.debug('trying %s, path %s' % (method,request.path))
        if method == 'session' and session is not None:
            # 'session.user' is set automtically using session cookie in request
            session_user = session.get('user', None)
            if session_user and (username is None or username == session_user):
                auth_user = session_user
                auth_date = session.get('auth_date', None)
                auth_method = method
                logger.debug('check_auth session: User %s authenticated using session cookie' % (auth_user,))

        elif method == 'passwd' and valid_username(username) and username.lower() != app.conf['wapt_user'].lower() and password is not None:
            # local htpasswd user/passwd file for add_host registration action
            htpasswd_users = get_htpasswdfile()
            if htpasswd_users.get_hash(username):
                if htpasswd_users.check_password(username, password):
                    auth_method = method
                    auth_user = username
                    auth_date = datetime.datetime.utcnow().isoformat()
                    logger.debug('check_auth passwd: User %s authenticated using htpasswd %s' % (auth_user, htpasswd_users))
                else:
                    logger.debug('check_auth passwd: User %s htpasswd %s verification failed' % (username, htpasswd_users))
                    time.sleep(3)

        elif method == 'admin' and valid_username(username) and app.conf['wapt_user'] == username and password is not None:
            pbkdf2_sha256_ok = False
            pass_sha512_crypt_ok = False
            pass_bcrypt_crypt_ok = False

            if '$pbkdf2-sha256$' in app.conf['wapt_password']:
                pbkdf2_sha256_ok = pbkdf2_sha256.verify(password, app.conf['wapt_password'])
            elif sha512_crypt.identify(app.conf['wapt_password']):
                pass_sha512_crypt_ok = sha512_crypt.verify(
                    password,
                    app.conf['wapt_password'])
            else:
                try:
                    if bcrypt.identify(app.conf['wapt_password']):
                        pass_bcrypt_crypt_ok = bcrypt.verify(
                            password,
                            app.conf['wapt_password'])
                except Exception:
                    pass

            if pbkdf2_sha256_ok or pass_sha512_crypt_ok or pass_bcrypt_crypt_ok:
                auth_method = method
                auth_user = username
                auth_date = datetime.datetime.utcnow().isoformat()
                logger.debug('check_auth admin: wapt admin user authenticated using embedded config password')
            else:
                logger.debug('check_auth admin: wapt admin passwd verification failed')
                time.sleep(3)

        elif method == 'token':
            # token auth
            token_secret_key = app.conf['token_secret_key']
            if token_secret_key:
                # check if there is a valid token in the password
                try:
                    token_gen = get_secured_token_generator(token_secret_key)
                    # check if there is a Bearer, else use Basic / password
                    authorization = request.headers.get('Authorization')
                    max_age = app.conf['token_lifetime']
                    if authorization and authorization.startswith('Bearer '):
                        token_data = token_gen.loads(authorization.split(' ', 1)[1], max_age=max_age)
                    elif password is not None:  # we accept to get the token in password location instead of the actual password.
                        try:
                            token_data = token_gen.loads(password)
                        except BadData:
                            raise EWaptAuthenticationFailure('No valid token in password')
                    else:
                        raise EWaptAuthenticationFailure('No token')

                    uuid = token_data.get('uuid', token_data.get('user', None))
                    if not uuid:
                        raise EWaptAuthenticationFailure('Bad token UUID')
                    if token_data['server_uuid'] != get_server_uuid():
                        raise EWaptAuthenticationFailure('Bad server UUID')
                    auth_method = method
                    auth_user = uuid
                    auth_date = token_data['auth_date']
                    logger.debug('check_auth token: User %s authenticated using token' % (uuid,))
                except Exception as e:
                    logger.debug('check_auth token: Token verification failed : %s' % repr(e))
                    time.sleep(3)

        elif method == 'kerb' and app.conf['use_kerberos'] and request.path.endswith('_kerberos'):
            # with nginx kerberos module, auth user name is stored as Basic auth in the
            # 'Authorisation' header with password 'bogus_auth_gss_passwd'
            logger.debug('check_auth kerb with User %s, request %s' % (username,request))
            # Kerberos auth negociated by nginx
            authorization = request.authorization
            if authorization:
                logger.debug(repr(authorization))
                if authorization.username != '' and authorization.password == 'bogus_auth_gss_passwd':
                    authenticated_user = authorization.username.lower()
                    auth_method = method
                    auth_user = authenticated_user
                    auth_date = datetime.datetime.utcnow().isoformat()
                    logger.debug('check_auth kerb: User %s authenticated using kerberos' % (authenticated_user,))
                else:
                    time.sleep(3)

        elif method == 'ldap' and app.conf['ldap_auth_server'] and valid_username(username) and password is not None and username.lower() != app.conf['wapt_user'].lower() and (not request.path.endswith('_kerberos')):
            logger.debug("check_auth ldap: Trying LDAP authentication for user %s " % username)
            auth_dict_output = auth_module_ad.check_credentials_ad(app.conf, username, password,
                                                   list_dn=[app.conf.get('wapt_admin_group_dn')],
                                                   list_group=app.conf.get('wapt_admin_group'))
            logger.debug(auth_dict_output)
            if auth_dict_output['error']==True:
                logger.error('check_auth ldap: An error occured when trying LDAP authentication : %s ' % auth_dict_output['msg'])
            if auth_dict_output['success']==False:
                logger.info('check_auth ldap: Authentication failed : %s ' % auth_dict_output['msg'])
            if auth_dict_output['groups']:
                auth_method = method
                auth_user = username
                auth_date = datetime.datetime.utcnow().isoformat()
                logger.debug('check_auth ldap : User %s authenticated using LDAP' % (username,))

        # nginx client ssl certificate auth. CN of certificate's subject must match the expected username
        elif method == 'ssl' and request and app.conf['use_ssl_client_auth']:
            dn_user = get_cn_from_x_ssl_client_auth(request)
            if dn_user:
                dn_user = dn_user.lower()
                # if username is provided, be sure it matches the CN to avoid misunderstanding
                if dn_user and (not username or (dn_user.lower() == username.lower())):
                    auth_method = method
                    auth_user = dn_user.lower()
                    auth_date = datetime.datetime.utcnow().isoformat()
                    logger.debug('check_auth: User %s authenticated using SSL client certificate' % (auth_user,))
                else:
                    logger.debug('check_auth: CN mismatch using SSL client certificate. user: %s CN from cert DN: %s' % (username,dn_user))

        if auth_method and auth_user:
            break

    # only session and admin methods are ok for the embedded wapt admin password based account.
    if auth_method and auth_user and (auth_user.lower() != app.conf['wapt_user'].lower() or auth_method in ('session', 'admin', 'token')):
        return dict(auth_method=auth_method, user=auth_user, auth_date=auth_date)
    else:
        return None


def change_admin_password(newpassword):
    """Computes and stores the hash of newpassword for the embedded wapt admin account
    in waptserver.ini config file.

    """
    new_hash = pbkdf2_sha256.hash(newpassword)
    rewrite_config_item(app.config['CONFIG_FILE'], 'options', 'wapt_password', new_hash)
    app.conf['wapt_password'] = new_hash
    return new_hash



# edit_xxxx_package is subject to the edit_packages list of package name patterns for actual authorization.
# if  edit_packages is NULL, there is no restriction on list of packages which can be edited (CRUD).
ALL_ACLS = ['admin', 'view',
        'register_host','unregister_host',
        'trigger_host_action','trigger_host_upgrade',
        'edit_repo_rules',
        'edit_host_package', 'edit_base_package', 'edit_group_package',
        'edit_unit_package', 'edit_wsus_package', 'edit_profile_package', 'edit_selfservice_package',
        'edit_report','run_report','assign_package',
        ]


def merge_acls_dicts(d1,d2):
    """Merge keys. values are lists. if key exists, append value to list
    >>> merge_acls_dicts({'view':False,'admin':True},{'view':['123','456'],'admin':None})
    {'admin': None, 'view': False}
    """
    if d2:
        for k2,v2 in d2.items():
            v1 = d1.get(k2)
            if v1 == True:
                # acl for any perimeter
                continue
            elif v2 == True or v1 is None or v1 == False:
                # acl for any perimeter or new acl
                d1[k2] = v2
            elif isinstance(v1,list):
                # append new perimeters to existing acl
                for p in v2:
                    if not p in v1:
                        d1[k2].append(p)
    return d1

def get_user_acls(username=None,user_fingerprint=None):
    """Return list of non expired acls for a user

    Args:
        username (str): filter on username
        user_fingerprint (str) : filter on user certificate sha256 fingerprint

    Returns:
        list of dict {'aclname':bool or list}
    """


    # acl with username = '*' apply to all users (default acls)
    must_close_on_exit = wapt_db.is_closed()
    if must_close_on_exit:
        wapt_db.connect()

    try:
        filter = (
            (WaptUserAcls.expiration_date.is_null() | (WaptUserAcls.expiration_date >= datetime2isodate(datetime.datetime.utcnow()))) &
            ~ WaptUserAcls.acls.is_null()
            )
        if username is not None:
            filter = filter & WaptUserAcls.username.in_([username,'*'])
        if user_fingerprint is not None:
            filter = filter & WaptUserAcls.user_fingerprint.in_([user_fingerprint,'*'])
        result = {}
        for r in WaptUserAcls.select(WaptUserAcls.acls).where(filter).tuples():
            merge_acls_dicts(result,r[0])
        return result
    finally:
        if must_close_on_exit:
            wapt_db.close()


def has_any_right(user_acls, any_of_acls=['admin'], perimeters=None):
    """Return the list of perimeters for which the user
    has any of the rights listed in acls.

    Args:
        user_acls (dict) : acls of the user {'aclname':list or bool}
        perimeters (list of str): check the right in these perimeters only.
                                 If None, returns all matches

    Returns:
        list of str: list of perimeters fingerprint (list of CA fingerprint)
                     if no right on any perimeter, return empty list.

    >>> user_acls = {'admin':False,'view':['perimeter1_hash','perimeter2_hash'],'edit_base_package':True}
    >>> has_any_right(user_acls, ['view'],['aa'])
    []
    >>> has_any_right(user_acls, ['view'])
    ['perimeter2_hash', 'perimeter1_hash']
    >>> has_any_right(user_acls, ['admin'])
    []
    >>> has_any_right(user_acls, ['edit_base_package','admin'])
    True
    >>> has_any_right(user_acls, ['edit_base_package'])
    True
    >>> has_any_right(user_acls, ['edit_base_package','admin'])
    True
    >>> has_any_right(user_acls, ['edit_base_package'])
    True
    >>> has_any_right(user_acls, ['viewa'])
    []
    >>> has_any_right(user_acls, ['admin'])
    []
    >>> has_any_right(user_acls, ['view'],['perimeter1_hash','perimeter4_hash'])
    ['perimeter1_hash']
    """
    result = []
    if user_acls:
        for aclname in any_of_acls:
            acl_perimeters = user_acls.get(aclname)
            if isinstance(acl_perimeters,list):
                if perimeters is None:
                    # we append all
                    result.extend(acl_perimeters)
                elif set(acl_perimeters) & set(perimeters):
                    result.extend(list(set(acl_perimeters) & set(perimeters)))
            elif acl_perimeters == True:
                # we can stop here...
                return True
    return list(set(result))
