#!/usr/bin/env python3
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2023  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##
from flask import request, Blueprint, session

import time
import zlib
import ujson
import datetime

#from waptserver.utils import *
#from waptserver.model import *

from waptserver.utils import logger
from peewee import SQL,fn
from waptserver.model import wapt_db,wapt_db_connect
from waptserver.model import WaptUsers,Plugins,IsoConfig,Normalization,ReportingQueries,ReportingSnapshots

from waptserver.common import make_response, make_response_from_exception
from waptserver.decorators import requires_auth, wapt_db_readonly, require_wapt_db, requires_enterprise, reporting_enable

try:
    # i18n
    from flask_babel import Babel
    from flask_babel import gettext
except ImportError:
    gettext = (lambda s: s)
    Babel = None

_ = gettext

enterprise = Blueprint('enterprise', __name__)


def prepare_request_data(request):
    if request.headers.get("Content-Encoding") != "gzip":
        return
    data = zlib.decompress(request.data)
    request.data = data


@enterprise.route('/api/v3/normalization_import_softwares', methods=['GET'])
@requires_auth(['admin', 'view'])
@requires_enterprise
@require_wapt_db
def normalization_import_softwares():
    sql = """
    INSERT INTO normalization ( original_name, key, banned, windows_update, created_on, updated_on )
    SELECT
	case WHEN (S.name is null) or (S.name = '') THEN S.key ELSE S.Name END as original_name,
	S.key               as key,
	false               as banned,
	false               as windows_update,
        current_timestamp   as created_on,
        current_timestamp   as updated_on
    FROM
	normalization N
	RIGHT join ( SELECT name, key  FROM hostsoftwares GROUP BY name, key ) S ON (N.original_name = S.name) AND (N.key = S.KEY)
    WHERE
	N.original_name is NULL;
    """
    start_time = time.time()
    with wapt_db.atomic() as trans:
        try:
            c = wapt_db.execute_sql(sql)
            r = make_response(result=c.rowcount,  success=True, request_time=time.time() - start_time)
            return r
        except Exception as e:
            trans.rollback()
            return make_response_from_exception(e)


@enterprise.route('/api/v3/normalization_filter', methods=['POST'])
@requires_auth(['admin', 'view'])
@requires_enterprise
@require_wapt_db
def normalization_filter():
    start_time = time.time()
    prepare_request_data(request)
    data = ujson.loads(request.data)

    limit = int(data.get("limit", 1000))

    try:
        filter_string = data.get("filter_string", "").lower()
        logger.debug("filter string = " + filter_string)
        if data.get("only_kbs", False):
            data = Normalization.select(Normalization.key, Normalization.normalized_name, Normalization.banned, Normalization.windows_update,
            SQL("""CASE WHEN (original_name is null) or (original_name = '') THEN key ELSE original_name END""").alias('original_name')).where(
            Normalization.windows_update == True)
        else:
            data = Normalization.select(Normalization.key, Normalization.normalized_name, Normalization.banned, Normalization.windows_update,
            SQL("""CASE WHEN (original_name is null) or (original_name = '') THEN key ELSE original_name END""").alias('original_name'))

        if len(filter_string) > 0:
            data = data.where(
                fn.LOWER(Normalization.original_name).contains(filter_string) |
                fn.LOWER(Normalization.key).contains(filter_string) |
                fn.LOWER(Normalization.normalized_name).contains(filter_string)
            )

        data = data.order_by(Normalization.original_name)
        data = data.limit(limit)
        data = list(data.dicts())

        r = make_response({'softwares' : data, 'softwares_count' : len(data)} , success=True, request_time=start_time - time.time())
        return r
    except Exception as e:
        return make_response_from_exception(e)


@enterprise.route('/api/v3/normalization_write', methods=['POST'])
@requires_auth(['admin', 'run_report'])
@requires_enterprise
@require_wapt_db
def normalization_write():
    start_time = time.time()
    prepare_request_data(request)
    data = ujson.loads(request.data)
    with wapt_db.atomic() as trans:
        try:
            for l in data:
                n = Normalization()
                n.original_name = l['original_name']
                n.key = l['key']
                n.normalized_name = l['normalized_name']
                n.banned = l['banned']
                n.windows_update = l['windows_update']
                n.save()
            trans.commit()
            data = [len(data)]
            r = make_response(data, msg='normalization_write', success=True, request_time=start_time - time.time())
            return r
        except Exception as e:
            trans.rollback()
            return make_response_from_exception(e)


@enterprise.route('/api/v3/reporting_exec', methods=['GET'])
@requires_auth(['admin', 'run_report'])
@reporting_enable
@requires_enterprise
@wapt_db_readonly
def reporting_exec():
    start_time = time.time()
    R = None
    with wapt_db.atomic():
        try:
            id = request.args.get("id", None)
            if id is None:
                raise Exception("Param id is recquired")

            query = ReportingQueries.select(ReportingQueries.query).where(ReportingQueries.id == id)
            query = list(query.dicts())
            if 0 == len(query):
                raise Exception("Query id %d not found" % (id))
            data = ReportingQueries.raw(query[0]["query"])
            data = list(data.dicts())
            R = make_response(data, success=True, request_time=time.time() - start_time)
        except Exception as e:
            R = make_response_from_exception(e)
    return R


@enterprise.route('/api/v3/reporting_exec_raw', methods=['GET', 'POST'])
@requires_auth(['admin', 'run_report'])
@reporting_enable
@requires_enterprise
@wapt_db_readonly
def reporting_exec_raw():
    start_time = time.time()
    prepare_request_data(request)
    R = None
    with wapt_db.atomic() as trans:
        try:
            query = ujson.loads(request.data)
            if not query:
                raise Exception("Param query is recquired")

            sql = query['query']
            if not sql:
                raise Exception("Param query is recquired")

            data = ReportingQueries.raw(sql)
            data = list(data.dicts())
            R = make_response(data, success=True, request_time=time.time() - start_time)
        except Exception as e:
            trans.rollback()
            R = make_response_from_exception(e)
    return R


@enterprise.route('/api/v3/reporting_list', methods=['GET'])
@requires_auth(['admin', 'run_report'])
@reporting_enable
@requires_enterprise
@wapt_db_readonly
def reporting_list():
    start_time = time.time()
    with wapt_db:
        try:
            data = ReportingQueries.select(ReportingQueries).order_by(ReportingQueries.name)
            data = list(data.dicts())
            return make_response(data, success=True, request_time=time.time() - start_time)
        except Exception as e:
            return make_response_from_exception(e)


@enterprise.route('/api/v3/reporting_save', methods=['GET', 'POST'])
@requires_auth(['admin', 'edit_report'])
@reporting_enable
@requires_enterprise
@require_wapt_db
def reporting_save():
    start_time = time.time()
    prepare_request_data(request)
    try:
        queries_list = ujson.loads(request.data)
        if not queries_list:
            raise Exception("reporting : No data supplied")
    except Exception as e:
        return make_response_from_exception(e)

    with wapt_db.atomic() as trans:
        try:
            data = {}
            for query in queries_list:
                rq = ReportingQueries()
                if "id" in query:
                    if query["id"] >= 0:
                        rq.id = query["id"]
                if "name" in query:
                    rq.name = query["name"]
                if "query" in query:
                    rq.query = query["query"]
                if "settings" in query:
                    rq.settings = query["settings"]
                if "as_host_uuid" in query:
                    rq.as_host_uuid = query["as_host_uuid"]
                if "as_package_uuid" in query:
                    rq.as_package_uuid = query["as_package_uuid"]
                if "as_kb_update_id" in query:
                    rq.as_kb_update_id = query["as_kb_update_id"]
                if "as_software_key" in query:
                    rq.as_software_key = query["as_software_key"]
                if "snapshot_name" in query:
                    rq.snapshot_name = query["snapshot_name"]
                if "snapshot_scheduling" in query:
                    rq.snapshot_scheduling = query["snapshot_scheduling"]
                if "snapshot_ttl" in query:
                    rq.snapshot_ttl = query["snapshot_ttl"]

                rq.save()
                data[query["id"]] = rq.id
            trans.commit()
            R = make_response(data, success=True, request_time=time.time() - start_time)
            return R
        except Exception as e:
            trans.rollback()
            return make_response_from_exception(e)


@enterprise.route('/api/v3/reporting_delete', methods=['GET', 'POST'])
@requires_auth(['admin', 'edit_report'])
@reporting_enable
@requires_enterprise
@require_wapt_db
def reporting_delete():
    start_time = time.time()
    prepare_request_data(request)
    try:
        id_list = ujson.loads(request.data)
        if not id_list:
            raise Exception("reporting : No data supplied")
    except Exception as e:
        return make_response_from_exception(e)

    with wapt_db.atomic() as trans:
        try:
            cnt = ReportingQueries.delete().where(ReportingQueries.id.in_(id_list)).execute()
            trans.commit()
            R = make_response(cnt, success=True, request_time=time.time() - start_time)
            return R
        except Exception as e:
            trans.rollback()
            return make_response_from_exception(e)


@enterprise.route('/api/v3/reporting_snapshot', methods=['GET', 'POST','DELETE'])
@requires_auth(['admin', 'run_report'])
@reporting_enable
@requires_enterprise
@require_wapt_db
def reporting_snapshot():
    start_time = time.time()
    with wapt_db.atomic() as trans:
        try:
            if request.method == 'GET':
                report_id = int(request.args.get('report_id'))
                report = ReportingQueries.get(report_id)
                data = report.snapshot()

                msg = 'Snapshot of "%s" saved as "%s": %s records saved' % (report.name,report.snapshot_name,len(data))

                R = make_response(data, msg = msg, success=True, request_time=time.time() - start_time)
                return R

            """
            elif request.method == 'POST':
                prepare_request_data(request)
                try:
                    id_list = ujson.loads(request.data)
                    if not id_list:
                        raise Exception("reporting : No data supplied")
                except Exception as e:
                    return make_response_from_exception(e)

                with wapt_db.atomic() as trans:
                    try:
                        cnt = ReportingSnapshots.delete().where(ReportingQueries.id.in_(id_list)).execute()
                        trans.commit()
                        R = make_response(cnt, success=True, request_time=time.time() - start_time)
                        return R
                    except Exception as e:
                        trans.rollback()
                        return make_response_from_exception(e)
            """
        except Exception as e:
            trans.rollback()
            return make_response_from_exception(e)
# Users

@enterprise.route('/api/v3/waptusers', methods=['GET', 'PUT'])
@requires_auth(['admin'])
@requires_enterprise
@wapt_db_readonly
def waptusers_get():
    start_time = time.time()
    if request.method == 'GET':
        try:
            data = list(WaptUsers.select().dicts())
            R = make_response(data, success=True, request_time=time.time() - start_time, msg='%s Wapt users' % len(data))
            return R
        except Exception as e:
            return make_response_from_exception(e)
    elif request.method == 'PUT':
        # we have a single delta to apply
        prepare_request_data(request)
        try:
            delta = ujson.loads(request.data)
            if not delta:
                raise Exception("No delta rows list supplied")
        except Exception as e:
            return make_response_from_exception(e)

        with wapt_db.atomic() as trans:
            try:
                result = []
                for row in delta:
                    id = row.get('id')
                    if id is not None and id >= 0:
                        # get existing
                        user_rec = WaptUsers.get(id=id)
                    else:
                        # create new
                        user_rec = WaptUsers()
                    for (k, v) in row.items():
                        if k != 'id' and hasattr(WaptUsers, k):
                            setattr(user_rec, k, v)
                    user_rec.save()
                    result.append(user_rec)
                trans.commit()
                R = make_response(result, success=True, request_time=time.time() - start_time)
                return R
            except Exception as e:
                trans.rollback()
                return make_response_from_exception(e)


@enterprise.route('/api/v3/waptusers', methods=['POST'])
@requires_auth(['admin'])
@requires_enterprise
@require_wapt_db
def waptusers_post():
    DataClass = WaptUsers
    start_time = time.time()
    prepare_request_data(request)
    try:
        changes = ujson.loads(request.data)
        if not changes:
            raise Exception("No changes data supplied")
    except Exception as e:
        return make_response_from_exception(e)

    with wapt_db.atomic() as trans:
        try:
            result = []
            for change in changes:
                update_type = change.get('update_type')
                id = change.get('id')
                new_values = change.get('new_values')

                # insert
                if update_type in (1, 2):
                    if id is not None and id >= 0:
                        # get existing
                        rec = DataClass.get(id=id)
                    else:
                        # create new
                        rec = DataClass(**new_values)
                    # init with new values
                    if new_values:
                        for (k, v) in new_values.items():
                            if k != 'id' and hasattr(DataClass, k):
                                setattr(rec, k, v)
                    new_id = rec.save()
                    result.append((id, new_id))
                # delete
                elif update_type == 3:
                    DataClass.delete_by_id(id)
                    result.append((id, None))

            trans.commit()
            R = make_response(result, success=True, request_time=time.time() - start_time)
            return R
        except Exception as e:
            trans.rollback()
            return make_response_from_exception(e)


@enterprise.route('/api/v3/waptusers_delete', methods=['POST'])
@requires_auth(['admin'])
@requires_enterprise
@require_wapt_db
def waptusers_delete():
    start_time = time.time()
    prepare_request_data(request)
    try:
        id_list = ujson.loads(request.data)
        if not id_list:
            raise Exception("No data supplied")
    except Exception as e:
        return make_response_from_exception(e)

    with wapt_db.atomic() as trans:
        try:
            cnt = WaptUsers.delete().where(WaptUsers.id.in_(id_list)).execute()
            trans.commit()
            return make_response(cnt, success=True, request_time=time.time() - start_time, msg='%s users deleted' % len(id_list))
        except Exception as e:
            trans.rollback()
            return make_response_from_exception(e)


@enterprise.route('/api/v3/host_plugins', methods=['GET'])
@enterprise.route('/api/v3/import_plugins', methods=['GET']) # back compat
@requires_auth(['admin'])
@requires_enterprise
@require_wapt_db
def hosts_plugins():
    wapt_db_connect()
    try:
        data = list(Plugins.select().dicts())
        return make_response(data, success=True)
    except Exception as e:
        return make_response_from_exception(e)

@enterprise.route('/api/v3/save_plugins', methods=['POST'])
@requires_auth(['admin'])
@requires_enterprise
@require_wapt_db
def save_plugins():
    start_time = time.time()
    prepare_request_data(request)
    try:
        plugin_list = ujson.loads(request.data)
        if not plugin_list:
            raise Exception("Plugins : No data supplied")
    except Exception as e:
        return make_response_from_exception(e)

    with wapt_db.atomic() as db:
        try:
            data = []
            for plugin in plugin_list:
                # updates automatically the existing data ??
                pl = Plugins(**{k:plugin[k] for k in plugin if hasattr(Plugins,k)})
                pl.save()
                data.append(dict(pl))
            db.commit()
            return make_response(data, success=True, request_time=time.time() - start_time)

        except Exception as e:
            db.rollback()
            return make_response_from_exception(e)

@enterprise.route('/api/v3/delete_plugin', methods=['GET','POST'])
@requires_auth(['admin'])
@requires_enterprise
@require_wapt_db
def delete_plugin():
    start_time = time.time()
    prepare_request_data(request)
    try:
        plugin_to_delete = ujson.loads(request.data)
        if not plugin_to_delete:
            raise Exception("Plugins : No data supplied")
    except Exception as e:
        return make_response_from_exception(e)

    with wapt_db.atomic() as db:
        try:
            cnt = Plugins.delete().where(Plugins.name.in_(plugin_to_delete)).execute()
            return make_response(cnt, success=True, request_time=time.time() - start_time)
        except Exception as e:
            db.rollback()
            return make_response_from_exception(e)

def update_or_insert(Model,delta_recs,user):
    with wapt_db.atomic() as db:
        result = {}
        for rec_data in delta_recs:
            # id column is managed by database, fields starting with '_' are calculated fields.
            data = {k: rec_data[k] for k in rec_data if k in Model._meta.columns and k != 'id' and not k.startswith('_')}
            # negative id are client side generated temporary IDs for new inserted records.
            if rec_data.get('id',-1) < 0:
                data['created_on'] = datetime.datetime.utcnow()
                data['created_by'] = user
                created = Model.insert(**data).returning(Model.id).execute()
                if rec_data.get('id'):
                    result[rec_data.get('id')] = created[0].id
            else:
                # data comes from an existing record.
                data['updated_on'] = datetime.datetime.utcnow()
                data['updated_by'] = user
                update_count = Model.update(**data).where(Model.id == rec_data['id']).execute()
                if update_count > 0:
                    result[rec_data.get('id')] = rec_data.get('id')
        db.commit()
    return result

@enterprise.route('/api/v3/save_iso_config', methods=['POST'])
@requires_auth(['admin', 'wads_admin'])
@requires_enterprise
@require_wapt_db
def save_iso_config():
    start_time = time.time()

    prepare_request_data(request)
    delta_recs = ujson.loads(request.data)

    try:
        result = update_or_insert(IsoConfig, delta_recs, session.get('user'))
        return make_response(result=result, success=True, request_time=time.time() - start_time)
    except Exception as e:
        return make_response_from_exception(e)
