#!/usr/bin/waptpython
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2023  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##
import os
import sys
import platform

if platform.system() == 'Windows':
    print("Unsupported on Windows. Run winsetup.py instead.")
    sys.exit(1)

if "__file__" in locals():
    sys.path.insert(0, os.path.realpath(os.path.join( os.path.dirname(os.path.realpath(__file__)),'..','..')))

from waptutils import __version__


import shutil
import dialog
import subprocess
import jinja2
import socket
import uuid
import re
import psutil
import datetime
import pwd as good_pwd
import grp
from waptserver.scripts import nginxparser
import glob
import hashlib
import json
import configparser

from pathlib import Path
from optparse import OptionParser
from passlib.hash import pbkdf2_sha256
from passlib import pwd
from waptpackage import WaptLocalRepo
import waptserver.config

from waptserver.config import update_conf_for_secure_mode, make_secret
from waptserver.common import update_all_crls, get_wapt_full_version
from waptserver.utils import get_wapt_exe_version

from waptutils import get_local_IPs

from waptcrypto import SSLCertificate, SSLPrivateKey, SSLCRL
from setuphelpers import type_debian, type_redhat, run, mkdirs, get_distrib_version, Version, get_disk_free_space
from waptutils import sha256_for_file

import waptlicences

try:
    wapt_root_dir = os.path.abspath(os.path.join(os.path.dirname(os.path.realpath(__file__)), '../..'))
except:
    wapt_root_dir = 'c:/tranquilit/wapt'

postconf = dialog.Dialog(dialog="dialog")
dh_filename = '/etc/ssl/certs/dhparam.pem'
quiet = False
keytab_file = '/etc/nginx/http-krb5.keytab'

def get_current_running_postgresql_version():
    version = run("sudo -iu postgres psql -t -c 'SHOW server_version;'").split('(',1)[0].strip()
    try:
        return Version(version)
    except:
        print('cannot check current running postgresql version, aborting')
        sys.exit(1)


if type_debian():
    PGSQL_SVC = 'postgresql'
    wapt_folder = '/var/www/wapt'
    NGINX_ACCOUNT = 'www-data'
    NGINX_GID = grp.getgrnam(NGINX_ACCOUNT).gr_gid

elif type_redhat():
    # version 7 or version 8
    distrib_major_version = get_distrib_version().split('.')[0]
    if distrib_major_version=='7':
        if get_current_running_postgresql_version() > Version('14.0'):
            PGSQL_SVC='postgresql-14'
        elif get_current_running_postgresql_version() > Version('9.6') and get_current_running_postgresql_version() < Version('9.7'):
            PGSQL_SVC='postgresql-9.6'
        else:
            print('unsupported postgresql version')
            sys.exit(1)
    elif distrib_major_version=='8' or distrib_major_version=='9':
        PGSQL_SVC = 'postgresql'
    else:
        print("unsupported OS version")
        sys.exit(1)
    wapt_folder = '/var/www/html/wapt'
    NGINX_ACCOUNT = 'nginx'
    NGINX_GID = grp.getgrnam(NGINX_ACCOUNT).gr_gid

else:
    print("not supported, exiting")
    sys.exit(1)
############ FUNCTIONS ############


def check_if_deb_installed(package_name):
    child = subprocess.Popen('/usr/bin/dpkg -l "%s"' % package_name, stdout=subprocess.PIPE, shell=True)
    child.communicate()
    return child.returncode == 0


def check_if_rpm_installed(package_name):
    child = subprocess.Popen('/usr/bin/rpm -q "%s"' % package_name, stdout=subprocess.PIPE, shell=True)
    child.communicate()
    return child.returncode == 0


def guess_fqdn():
    """ Guess FQDN from hostname - return default value if nothing found"""
    try:
        fqdn = socket.getfqdn()
        if not fqdn:
            fqdn = 'wapt'
        if '.' not in fqdn:
            fqdn += '.lan'
    except:
        fqdn = 'srvwapt'
    return fqdn


def selinux_rules():
    """ SELinux httpd security rules """
    run('setsebool -P httpd_can_network_connect 1')
    run('setsebool -P httpd_setrlimit on')
    for sepath in ('wapt', 'wapt-host', 'waptwua','wads'):
        try:
            run('semanage fcontext -a -t httpd_sys_content_t "/var/www/html/%s(/.*)?"' % sepath)
        except:
            pass
        run('restorecon -R -v /var/www/html/%s' % sepath)


def setup_firewall():
    """ Add permanent rules for firewalld """
    if not type_redhat():
        # only firewalld is supported right now
        return

    # check if firewalld is running
    if run('firewall-cmd --state',accept_returncodes=[0,252]).strip()=='running':
        run('firewall-cmd --permanent --add-service=http')
        run('firewall-cmd --permanent --add-service=https')
        run('firewall-cmd --permanent --add-service=tftp')
        run('firewall-cmd --reload')
    else:
        run('firewall-offline-cmd --add-service=http')
        run('firewall-offline-cmd --add-service=https')
        run('firewall-offline-cmd --add-service=tftp')


#### WAPTSERVER ####


def enable_waptserver():
    out = run('systemctl enable waptserver')
    out += run('systemctl enable wapttasks')
    return out


def service_is_active(service):
    try:
        res = run('systemctl is-active %s' % service)
        return res.splitlines()[0] == 'active'
    except:
        return False

def service_is_enabled(service):
    try:
        res = run('systemctl is-enabled %s' % service)
        return res.splitlines()[0] == 'enabled'
    except:
        return False


def start_waptserver():
    out = run("systemctl restart waptserver")
    if service_is_enabled('waptserver-uwsgi'):
        out += run("systemctl restart waptserver-uwsgi")
    if service_is_enabled('wapttasks'):
        out += run("systemctl restart wapttasks")
    if service_is_enabled('wapttftpserver'):
        out += run("systemctl restart wapttftpserver")

    return out

UWSGI="""master = true
processes = 10
wsgi = waptserver.server:app
home = /opt/wapt
chdir = /opt/wapt
max-requests = 1000
socket = /run/waptserver/uwsgi.sock
uid = wapt
gid = www-data
chmod-socket = 664
env = CONFIG_FILE=/opt/wapt/conf/waptserver.ini
log-5xx = true
disable-logging = true
manage-script-name = true
mount = /=waptserver.server:app
"""

def append_uwsgi_config(cfgfile):
    # read configuration from waptserver.ini
    config = configparser.RawConfigParser()
    if os.path.exists(cfgfile):
        config.read(cfgfile)

    if not config.has_section('uwsgi'):
        config.add_section('uwsgi')

        for line in UWSGI.splitlines():
            k,v = line.split('=',1)
            config.set('uwsgi',k.strip(),v.strip())

        config.write(open(cfgfile,'w'))


#### NGINX ####


def generate_dhparam(key_size=2048):
    out = ''
    if not os.path.exists(dh_filename):
        out += run('openssl dhparam -out %s  %s' % (dh_filename, key_size))
    os.chown(dh_filename, 0, NGINX_GID)  # pylint: disable=no-member
    os.chmod(dh_filename, 0o640)  # pylint: disable=no-member
    return out

def make_nginx_config(waptserver_root_dir, fqdn, force_https, server_config, quiet=False, server_names=None):
    global wapt_root_dir

    ssl_dir = os.path.join(waptserver_root_dir, 'ssl')
    scripts_dir = os.path.join(waptserver_root_dir, 'scripts')
    wapt_ssl_key_file = os.path.join(ssl_dir, 'key.pem')
    wapt_ssl_cert_file = os.path.join(ssl_dir, 'cert.pem')
    mkdirs(ssl_dir)


    ipaddresses = [i for i in get_local_IPs() if not i.startswith('fe80::')]
    server_names = server_names or []

    if server_names:
        cn = server_names[0]
    else:
        cn = fqdn

    if server_names:
        altnames = server_names
    else:
        altnames = [cn,ipaddresses[0]]

    # write the nginx configuration fragment
    jinja_env = jinja2.Environment(loader=jinja2.FileSystemLoader(scripts_dir))
    template = jinja_env.get_template('wapt.nginxconfig.j2')
    krb5_realm = '.'.join(fqdn.split('.')[1:]).upper()

    template_vars = {
        'server_names': altnames,
        'waptserver_port': server_config['waptserver_port'],
        'wapt_repository_path': os.path.dirname(server_config['wapt_folder']),
        'windows': False,
        'debian': type_debian(),
        'redhat': type_redhat(),
        'force_https': force_https,
        'wapt_ssl_key_file': wapt_ssl_key_file,
        'wapt_ssl_cert_file': wapt_ssl_cert_file,
        'fqdn': fqdn,
        'use_kerberos': server_config.get('use_kerberos', False),
        'KRB5_REALM': krb5_realm,
        'wapt_root_dir': wapt_root_dir,
        'log_dir': '/var/log/nginx',
        'use_uwsgi': server_config.get('use_uwsgi', False),
        'use_ssl_client_auth': server_config.get('use_ssl_client_auth', False),
        'clients_signing_certificate': server_config.get('clients_signing_certificate'),
        'clients_signing_crl': server_config.get('clients_signing_crl', None),
        'known_certificates_folder': server_config.get('known_certificates_folder', None),
        'ssl_client_crls': server_config.get('ssl_client_crls', None),
        'ssl_client_certificates': server_config.get('ssl_client_certificates', None),
        'htpasswd_path': server_config.get('htpasswd_path', None),
        'wapt_dhparam_file' : dh_filename,
        'has_dhparam':True,
        'has_ipv6':waptserver.config.is_server_ipv6_capable(),
        'nginx_http': server_config['nginx_http'],
        'nginx_https': server_config['nginx_https'],
        'wads_enable': server_config.get('wads_enable'),
        'waptwua_enable': server_config.get('waptwua_enable'),
        'trusted_wads_subnets': server_config.get('trusted_wads_subnets'),
        'login_on_wads': server_config.get('login_on_wads'),
        'list_subnet_skip_login_wads': server_config.get('list_subnet_skip_login_wads'),
        'homepage_enable' : server_config['homepage_enable']
    }

    if quiet:
        print('[*] Nginx - creating wapt.conf virtualhost')

    config_string = template.render(template_vars)
    if type_debian():
        sites_available_path = '/etc/nginx/sites-available'
        sites_enabled_path = '/etc/nginx/sites-enabled'
        if not(os.path.isdir(sites_available_path)):
            mkdirs(sites_available_path)
        if not(os.path.isdir(sites_enabled_path)):
            mkdirs(sites_enabled_path)
        sites_available_path_wapt_conf = os.path.join(sites_available_path, 'wapt.conf')
        sites_enabled_path_wapt_conf = os.path.join(sites_enabled_path, 'wapt.conf')
        dst_file_path = sites_available_path_wapt_conf
        if not os.path.exists(sites_enabled_path_wapt_conf):
            os.symlink(sites_available_path_wapt_conf, sites_enabled_path_wapt_conf)
        else:
            if not os.path.islink(sites_enabled_path_wapt_conf):
                shutil.move(sites_enabled_path_wapt_conf, sites_enabled_path_wapt_conf + "." + str(datetime.datetime.utcnow().timestamp()) + '.disabled')
                os.symlink(sites_available_path_wapt_conf, sites_enabled_path_wapt_conf)
        sites_enabled_path_default = os.path.join(sites_enabled_path, 'default')
        if os.path.exists(sites_enabled_path_default):
            os.unlink(sites_enabled_path_default)

    elif type_redhat():
        dst_file_path = '/etc/nginx/conf.d/wapt.conf'
    # make a backup of current wapt nginx config.
    if os.path.exists(dst_file_path):
        with open(dst_file_path,'r') as f:
            if config_string != f.read():
                #
                backups_dir = os.path.join(os.path.dirname(os.path.dirname(dst_file_path)),'backups')
                if not os.path.isdir(backups_dir):
                    os.makedirs(backups_dir)
                shutil.move(dst_file_path,os.path.join(backups_dir,os.path.basename(dst_file_path) + "." + datetime.datetime.utcnow().strftime('%Y%m%d-%H%M%S') + '.backup'))

    with open(dst_file_path, 'wt') as dst_file:
        dst_file.write(config_string)

    # create keys for https:// access
    if not os.path.exists(wapt_ssl_key_file) or \
            not os.path.exists(wapt_ssl_cert_file):
        if quiet:
            print('[*] Nginx - generate self-signed certs')

        key = SSLPrivateKey(wapt_ssl_key_file)
        if not os.path.isfile(wapt_ssl_key_file):
            print('Create SSL RSA Key %s' % wapt_ssl_key_file)
            key.create()
            key.save_as_pem()

        if os.path.isfile(wapt_ssl_cert_file):
            crt = SSLCertificate(wapt_ssl_cert_file)
            if crt.cn != cn:
                shutil.move(wapt_ssl_cert_file, "%s-%s.old" % (wapt_ssl_cert_file, '{:%Y%m%d-%Hh%Mm%Ss}'.format(datetime.datetime.utcnow())))
                crt = key.build_sign_certificate(
                    cn=cn,
                    altnames=altnames,
                    is_server_auth=True,
                    is_client_auth=False,
                    is_code_signing=False,
                    is_ca=False,
                    key_usages=['digital_signature'],
                    validity_days=server_config['server_certificate_lifetime']
                )
                print('Create X509 cert %s' % wapt_ssl_cert_file)
                crt.save_as_pem(wapt_ssl_cert_file)
        else:
            crt = key.build_sign_certificate(
                cn=cn,
                altnames=altnames,
                is_server_auth=True,
                is_client_auth=False,
                is_code_signing=False,
                is_ca=False,
                key_usages=['digital_signature'],
                validity_days=server_config['server_certificate_lifetime']
            )
            print('Create X509 cert %s' % wapt_ssl_cert_file)
            crt.save_as_pem(wapt_ssl_cert_file)
    else:
        if quiet:
            print('[*] Nginx - self-signed certs already exists, skipping...')


def enable_nginx():
    return run('systemctl enable nginx')


def restart_nginx():
    return run('systemctl restart nginx')


def nginx_set_worker_limit(nginx_conf):
    already_set = False
    for entries in nginx_conf:
        if entries[0] == 'worker_rlimit_nofile':
            print("[*] Nginx - worker_rlimit_nofile already set")
            already_set = True
    if not already_set:
        nginx_conf.insert(3, ['worker_rlimit_nofile', '32768'])
    return nginx_conf


def nginx_set_worker_connections(nginx_conf):
    for entry in nginx_conf:
        if entry[0] == ['events']:
            for subentry in entry[1]:
                if subentry[0] == 'worker_connections':
                    if subentry[1] != '4096':
                        subentry[1] = '4096'
                        return nginx_conf
    return nginx_conf


def nginx_clean_default_vhost(nginx_conf):
    for entry in nginx_conf:
        if entry[0] == ['http']:
            found_gzip = False
            for subentry in entry[1]:
                if subentry[0] == ['server']:
                    print('[*] Nginx - removing default vhost')
                    entry[1].remove(subentry)
                if subentry[0] == 'gzip':
                    if subentry[1] == 'on':
                        found_gzip = True
                    else:
                        found_gzip = True
                        subentry[1] = 'on'
            if not(found_gzip):
                entry[1].insert(1, ['gzip', 'on'])
    return nginx_conf


def nginx_cleanup():
    with open('/etc/nginx/nginx.conf', 'r') as read_conf:
        nginx_conf = nginxparser.load(read_conf)
        nginx_conf = nginx_set_worker_limit(nginx_conf)
        nginx_conf = nginx_set_worker_connections(nginx_conf)
        nginx_conf = nginx_clean_default_vhost(nginx_conf)
    with open("/etc/nginx/nginx.conf", "w") as nginx_conf_file:
        nginx_conf_file.write(nginxparser.dumps(nginx_conf))


def get_folder_size(root_directory):
    root_directory=Path(root_directory)
    return sum(f.stat().st_size for f in root_directory.glob('**/*') if f.is_file())


#### POSTGRESQL ####
def migrate_db_if_needed():

     if get_current_running_postgresql_version() < Version('10'):

        if type_redhat():
            print('migration to postgresql 14 required')
            print('Init db pgsql-14')
            run('/usr/pgsql-14/bin/postgresql-14-setup initdb', cwd='/opt/wapt')
            print('stop db pgsql-9.6')
            run('systemctl stop postgresql-9.6')
            try:
                run('sudo -u postgres psql wapt -c "DROP EXTENSION hstore;"')
            except:
                pass
            print('pg_upgrade running')

            #Do we have enough disk space?
            old_database_folder= '/var/lib/pgsql/9.6/data'
            new_database_folder= '/var/lib/pgsql/14/data'

            free_space = int(get_disk_free_space(new_database_folder))
            free_space_delta = free_space - int(get_folder_size(old_database_folder))
            # check if the remaining disk space available supports the size of the current database
            if free_space_delta < 1024*1024*1024 :
                print('not enough disk space available for migration to postgresql 14')
                sys.exit(1)

            run('sudo -u postgres /usr/pgsql-14/bin/pg_upgrade --old-bindir=/usr/pgsql-9.6/bin/ --new-bindir=/usr/pgsql-14/bin/ --old-datadir=%s/ --new-datadir=%s/' % (old_database_folder,new_database_folder),cwd='/var/lib/pgsql/14/data/')
            print('disable postgresql-9.6')
            run('systemctl disable postgresql-9.6')
            print('enable postgresql-14')
            run('systemctl enable postgresql-14')
            run('systemctl start  postgresql-14')
            global PGSQL_SVC
            PGSQL_SVC='postgresql-14'
        else:

            #retrieve all clusters present
            data=json.loads(run('pg_lsclusters --json'))

            old_database=[u for u in data if u['port'] == "5432"][0]
            new_database = sorted(data,key=lambda p:(Version(p['version'])),reverse=True)[0]

            old_database_folder=old_database['config']['data_directory']
            new_database_folder=new_database['config']['data_directory']

            old_database_version=old_database['version']
            new_database_version = new_database['version']

            #We must be in 10 minimum
            if Version(new_database_version) < Version('10'):
                print('Error last version postgresql found %s < 10' % new_database_version)
                sys.exit(1)

            #Do we have enough disk space?
            free_space = int(get_disk_free_space(new_database_folder))
            free_space_delta = free_space - int(get_folder_size(old_database_folder))
            # check if the remaining disk space available supports the size of the current database
            if free_space_delta < 1024*1024*1024 :
                print('not enough disk space available for migration to postgresql %s' % new_database_version)
                sys.exit(1)

            # Is the new database empty?
            list_base = [ u.strip() for u in run("""sudo -iu postgres psql -p %s -t -c "SELECT datname FROM pg_database;" """ % new_database['port']).split('\n') if u.strip()]
            unexcepted_database= [u.lower() for u in list_base if not u in ['template0','template1','postgres']]
            if unexcepted_database:
                print('Error Migration to postgresql %s needed but %s not empty, database found %s' % (new_database_version,new_database_version,' '.join(list_base)))
                sys.exit(1)
            #Ok we migrate
            print('delete empty database %s' % new_database_version)
            run("su - postgres -c 'pg_dropcluster --stop %s main'" % new_database_version)
            print('migrate database %s to %s' % (old_database_version,new_database_version))
            run("su - postgres -c 'pg_upgradecluster %s main'" % old_database_version)
            print("successful migration")
            print("""you can run for delete old database "su - postgres -c 'pg_dropcluster --stop %s main'"  """ % old_database_version)


def ensure_postgresql_db(db_name='wapt', db_owner='wapt', db_password=''):
    """ create postgresql wapt db and user if it does not exists """

    migrate_db_if_needed()

    def postgresql_running():
        return [p for p in psutil.process_iter() if p.name().lower() in (PGSQL_SVC)]

    if not postgresql_running():
        run('systemctl start %s' % PGSQL_SVC)
        run('systemctl enable %s' % PGSQL_SVC)

    val = run(""" sudo -u postgres psql template1 -c " select usename from pg_catalog.pg_user where usename='%s';"  """ % db_owner, cwd='/')
    if 'wapt' in val:
        print("[*] postgresql - user wapt already exists, skipping creating user  ")
    else:
        print("[*] postgresql - we suppose that the db does not exist either (or the installation has been screwed up)")
        if not db_password:
            run(""" sudo -u postgres psql template1 -c "create user %s ; " """ % (db_owner), cwd='/')
        else:
            run(""" sudo -u postgres psql template1 -c "create user %s with password '%s'; " """ % (db_owner, db_password), cwd='/')

    val = run(""" sudo -u postgres psql template1 -c " SELECT datname FROM pg_database WHERE datname='%s';   " """ % db_name, cwd='/')

    if 'wapt' in val:
        print("[*] postgresql - db already exists, skipping db creation")
    else:
        print('[*] postgresql - creating db wapt')
        run(""" sudo -u postgres psql template1 -c "create database %s with owner=%s encoding='utf-8'; " """ % (db_name, db_owner), cwd='/')

    # check if hstore (for json) is installed
    val = run(""" sudo -u postgres psql wapt -c "select * from pg_extension where extname='hstore';" """, cwd='/')
    if 'hstore' in val:
        print("[*] postgresql - hstore extension already loading into database, skipping create extension")
    else:
        run(""" sudo -u postgres psql wapt -c "CREATE EXTENSION hstore;" """, cwd='/')

    # check if jsquery (for json) is installed
    val = run(""" sudo -u postgres psql wapt -c "select * from pg_extension where extname='jsquery';" """, cwd='/')
    if 'jsquery' in val:
        print("[*] postgresql - jsquery extension already loading into database, skipping create extension")
    else:
        try:
            run(""" sudo -u postgres psql wapt -c "CREATE EXTENSION jsquery;" """, cwd='/')
        except:
            print('[ ] postgresql jsquery extension is not available. This is not mandatory. Check if a package like postgresql-11-jsquery is available for your distribution')


def generate_keytab():
    if type_debian():
        if check_if_deb_installed('krb5-user msktutil libnginx-mod-http-auth-spnego'):
            print('Please install krb5-user msktutil libnginx-mod-http-auth-spnego')

    if type_redhat():
        if check_if_rpm_installed('krb5-workstation msktutil nginx-mod-http-auth-spnego'):
            print('Please install krb5-workstation msktutil nginx-mod-http-auth-spnego')

    fqdn = guess_fqdn()
    krb5_realm = '.'.join(fqdn.split('.')[1:]).upper()

    msg = "Enter Kerberos REALM"
    (code, reply) = postconf.inputbox(text=msg, width=50, init=krb5_realm)
    if code != postconf.DIALOG_OK:
        exit(1)
    else:
        krb5_realm = reply

    if os.path.isfile('/etc/krb5.conf'):
        os.remove('/etc/krb5.conf')

    krb5_config = """[libdefaults]
    default_realm = %s
    dns_lookup_kdc = true
    dns_lookup_realm = false
""" % krb5_realm

    with open('/etc/krb5.conf','w') as krb5_configfile:
        krb5_configfile.write(krb5_config)

    msg = "Enter a Domain Controller name in write mode"
    (code, reply) = postconf.inputbox(text=msg, width=50)
    if code != postconf.DIALOG_OK:
        exit(1)
    else:
        dc_name = reply.upper()

    msg = "Enter a username with administrator privileges"
    (code, reply) = postconf.inputbox(text=msg, width=50, init="administrator")
    if code != postconf.DIALOG_OK:
        exit(1)
    else:
        admin_username = reply


    msg = "Enter administrator password"
    (code, reply) = postconf.passwordbox(text=msg, insecure=True, width=50)
    if code != postconf.DIALOG_OK:
        exit(1)
    else:
        admin_passwd = reply

    try:
        run("echo '%s' | /usr/bin/kinit %s" % (admin_passwd,admin_username))
        run('/usr/bin/msktutil --server %s --precreate --host %s -b cn=computers --service HTTP --description "host account for wapt server" --enctypes 24 -N' % (dc_name,fqdn))
        run('/usr/bin/msktutil --server %s --auto-update --keytab /etc/nginx/http-krb5.keytab --host %s -N' %(dc_name,fqdn))
        postconf.msgbox('Keytab file is correctly generated in %s' % keytab_file)
    except:
        postconf.msgbox('Unable to create keytab file. Please refert to \nhttps://www.wapt.fr/en/doc/wapt-security-configuration-server.html#configuring-kerberos-authentication', width=120)

    if os.path.isfile(keytab_file):
        if type_debian():
            run("/bin/chmod 640 %s" % keytab_file)
            run("/bin/chown root:www-data /etc/nginx/http-krb5.keytab")

        if type_redhat():
            run("/bin/chmod 640 /etc/nginx/http-krb5.keytab")
            run("/bin/chown root:nginx /etc/nginx/http-krb5.keytab")
            run("/usr/sbin/restorecon -v -R /etc/nginx/http-krb5.keytab")


def configure_max_connections():
    config_file_path = run(""" sudo -iu postgres psql -t -c 'SHOW config_file;' """).strip()
    print("[*] postgresql changing max_connections parameter in postgres config file %s" % config_file_path)
    with open(config_file_path, "r+") as config_file:
        new_lines = []
        found = False
        for line in config_file.readlines():
            if 'max_connections' in line:
                new_lines.append('max_connections = 1000\n')
                found = True
            else:
                new_lines.append(line)
        if not(found):
            new_lines.append('max_connections = 1000\n')
        config_file.seek(0)
        config_file.truncate()
        config_file.writelines(new_lines)

def main():
    global wapt_root_dir

    usage = """%prog [--config filename] [--force-https] [--quiet] [--dhparam-key-size=SIZE]"""

    parser = OptionParser(usage=usage, version=__version__)
    parser.add_option(
        '-c',
        '--config',
        dest='configfile',
        default=waptserver.config.DEFAULT_CONFIG_FILE,
        help='Config file full path (default: %default)')
    parser.add_option(
        "-s",
        "--force-https",
        dest="force_https",
        default=False,
        action='store_true',
        help="Use https only, http is 301 redirected to https (default: False). Requires a proper DNS name")
    parser.add_option(
        '-q',
        '--quiet',
        dest='quiet',
        default=False,
        action="store_true",
        help='Run quiet postconfiguration - default password and simple behavior')
    parser.add_option(
        '-n',
        '--nginx',
        dest='nginx',
        default=False,
        action="store_true",
        help='Run nginx configuration if in quiet mode')
    parser.add_option(
        '--dhparam-key-size',
        dest='dhparam_key_size',
        default=2048,
        metavar='NUMBER',
        type='int',
        help='Size for dhparam key')
    parser.add_option(
        '-p',
        '--admin-password',
        dest='password',
        default=None,
        help='Password for waptserver')
    parser.add_option(
        '--waptbasedir',
        dest='waptbasedir',
        default=wapt_root_dir,
        help='Wapt installation base dir (default: %default)')
    parser.add_option(
        '--server-names',
        dest='server_names',
        default=None,
        help='server name and ip for certificate CN and SubjectAltNames. Separator is a comma (default: %default)')
    parser.add_option(
        '--cspn-toe',
        dest='secured_mode',
        default=False,
        action='store_true',
        help="""Define settings for CSPN TOE mode (not recommended for most use cases) (default: %default)
 - Disable WADS, WAPTWUA, reporting, secondary repositories, repo rules, server home page.
 - Disable auto create of ldap users.
 - Enforce redirect to https
 - Enforce Kerberos
 - Enforce use of CRL for agents certificates
 - Disable ldap and passwd login methods. Login only with 'kerb' and 'admin' methods.
 - Authenticate API endpoints only with 'session' method
 - Agent registration authentication restricted to 'admin', 'passwd','kerb'
 - Enforce long (>=20) and complex passwords.
 - Reduced default lifetime of certificates
 - Reduced lifetime of session
""")
    parser.add_option(
        '--use-uwsgi',
        dest='use_uwsgi',
        default=False,
        action='store_true',
        help='Configure a waptserver-uwsgi service in addition to waptserver (default: %default)')

    (options, args) = parser.parse_args()
    wapt_root_dir = options.waptbasedir

    global quiet
    quiet = options.quiet

    nginx = options.nginx

    if options.server_names is not None:
        server_names = [s.strip() for s in options.server_names.split(',')]
    else:
        server_names = None

    global dhparam_key_size
    dhparam_key_size = options.dhparam_key_size

    def message_or_print(message):
        if quiet:
            print('[*] '+message)
        else:
            postconf.msgbox(message)

    try:
        # Load existing config file
        server_config = waptserver.config.load_config(options.configfile)

        if options.use_uwsgi:
            server_config['use_uwsgi'] = True

        if not quiet:
            if postconf.yesno("Do you want to launch post configuration tool ?") != postconf.DIALOG_OK:
                print("Canceling wapt postconfiguration")
                sys.exit(1)
        else:
            print('WAPT silent post-configuration')

        # SELinux rules for CentOS/RedHat
        if type_redhat():
            if re.match('^SELinux status:.*enabled', run('sestatus')):
                message_or_print('SELinux detected, tweaking httpd permissions.')
                selinux_rules()
                message_or_print('SELinux correctly configured for Nginx reverse proxy')

        # update settings to have minimum requirements for CSPN TOE
        if options.secured_mode:
            options.force_https = True
            update_conf_for_secure_mode(server_config, wapt_root_dir = wapt_root_dir)

        if os.path.isfile(options.configfile):
            print('[*] waptserver Making a backup copy of the configuration file')
            datetime_now = datetime.datetime.utcnow()
            shutil.copyfile(options.configfile, '%s.bck_%s' % (options.configfile, datetime_now.isoformat()))

        global wapt_folder

        wapt_folder = server_config['wapt_folder']

        # add secret key initialisation string (for session cookies)
        if not server_config['secret_key']:
            server_config['secret_key'] = make_secret()

        # add secret key initialisation string (for token)
        if not server_config['token_secret_key']:
            server_config['token_secret_key'] = make_secret()

        # add user db and password in ini file
        if server_config['db_host'] in (None, '', 'localhost', '127.0.0.1', '::1'):
            ensure_postgresql_db(db_name=server_config['db_name'], db_owner=server_config['db_name'], db_password=server_config['db_password'])
            print("Upgrading DB structure.")
            run('sudo -u wapt bash --noprofile --norc /opt/wapt/waptpython.sh /opt/wapt/waptserver/model.py upgrade-db -l info -c "%s"' % options.configfile)

        configure_max_connections()

        # Password setup/reset screen
        if not quiet:
            current_password_hash = server_config['wapt_password']
            # check if old hash
            if not current_password_hash or \
               not '$pbkdf2' in current_password_hash or \
                    postconf.yesno("Do you want to reset main admin (user: %s) password ?" % server_config['wapt_user'], yes_label='skip', no_label='reset') != postconf.DIALOG_OK:
                wapt_password_ok = False
                while not wapt_password_ok:
                    wapt_password = ''
                    wapt_password_check = ''

                    while wapt_password == '':
                        if options.secured_mode:
                            msg = "Please enter the wapt server password (min. %s characters, punctuation, upper and lower case):" % server_config['min_password_length']
                        else:
                            msg = "Please enter the wapt server password (min. %s):" % server_config['min_password_length']
                        (code, wapt_password) = postconf.passwordbox(msg, insecure=True, width=100)
                        if code != postconf.DIALOG_OK:
                            exit(0)

                    while wapt_password_check == '':
                        (code, wapt_password_check) = postconf.passwordbox("Please enter the wapt server password again:  ", insecure=True, width=100)
                        if code != postconf.DIALOG_OK:
                            exit(0)

                    if wapt_password != wapt_password_check:
                        postconf.msgbox('Password mismatch !')
                    else:
                        password_error = waptlicences.check_password_complexity(wapt_password, server_config.get('check_password_complexity',False))
                        if password_error:
                            postconf.msgbox('error: %s' % password_error)
                        else:
                            wapt_password_ok = True

                password = pbkdf2_sha256.hash(wapt_password.encode('utf8'))
                server_config['wapt_password'] = password
        else:
            wapt_password = ''
            if not server_config['wapt_password']:
                print('[*] Generating random password for WAPT server')
                wapt_password = options.password if options.password else pwd.genword(entropy=56, charset="ascii_62")
                print('[*] WAPT admin password : %s' % wapt_password)
                password = pbkdf2_sha256.hash(wapt_password.encode('utf8'))
                server_config['wapt_password'] = password

        if not server_config['server_uuid']:
            server_config['server_uuid'] = str(uuid.uuid1())


        # waptagent authentication method
        if not quiet:
            # in secured mode, we force kerberos and authentication on websockets
            if not options.secured_mode:
                defaultConfigKerb : bool = server_config.get('use_kerberos')
                defaultConfigRegistration : bool = server_config.get('allow_unauthenticated_registration')
                # if it's the first launch we propose allow_unauthenticated_registration = true
                if not os.path.isfile(options.configfile) :
                    defaultConfigRegistration = True
                choices = [
                    ("1", "Allow unauthenticated registration", (not defaultConfigKerb) and defaultConfigRegistration),
                    ("2", "Enable kerberos authentication required for machines registration", (defaultConfigKerb and not defaultConfigRegistration)),
                    ("3", "Disable Kerberos but registration require strong authentication", not defaultConfigKerb and not defaultConfigRegistration),
                ]

                code, t = postconf.radiolist("WaptAgent Authentication type?", choices=choices, width=120)
                if code == 'cancel':
                    print("\n\npostconfiguration canceled\n\n")
                    sys.exit(1)
                if t == "1":
                    server_config['allow_unauthenticated_registration'] = True
                    server_config['use_kerberos'] = False
                if t == "2":
                    server_config['allow_unauthenticated_registration'] = False
                    server_config['use_kerberos'] = True
                    if not os.path.isfile(keytab_file):
                        generate_keytab()

                if t == "3":
                    server_config['allow_unauthenticated_registration'] = False
                    server_config['use_kerberos'] = False
        else:
            print('[*] Set default registration method to : Allow anyone to register + Kerberos disabled')
            server_config['allow_unauthenticated_registration'] = True
            if server_config['use_kerberos'] is None:
                server_config['use_kerberos'] = False

        # https client auth
        if not quiet:
            if not options.secured_mode:
                choices = [
                    ("1", "Authenticate Agents using https client certificate (recommended)", server_config['use_ssl_client_auth']),
                    ("2", "Don't check https client certificate (legacy)", not server_config['use_ssl_client_auth']),
                ]

                code, t = postconf.radiolist("WaptAgent client certificate checking", choices=choices, width=120)
                if code == 'cancel':
                    print("\n\npostconfiguration canceled\n\n")
                    sys.exit(1)
                if t == "1":
                    server_config['use_ssl_client_auth'] = True
                    server_config['allow_unauthenticated_connect'] = False
                if t == "2":
                    server_config['use_ssl_client_auth'] = False
                    server_config['allow_unauthenticated_connect'] = True # we don't have websocket token anymore
            else:
                server_config['use_ssl_client_auth'] = True

        # WADS
        if not quiet:
            if not options.secured_mode:
                if postconf.yesno("Do you want to activate os deployment ?") == postconf.DIALOG_OK:
                    server_config['wads_enable'] = True
                else:
                    server_config['wads_enable'] = False

                if server_config['wads_enable']:
                    if postconf.yesno("Would you like to activate secure authentication on wads ?") == postconf.DIALOG_OK:
                        server_config['login_on_wads'] = True
                    else:
                        server_config['login_on_wads'] = False

                    if server_config['login_on_wads']:
                        if postconf.yesno("Would you like to mention subnet ip exempt from wads authentication") == postconf.DIALOG_OK:
                            msg= 'Mention here the subnets ex: 192.168.1.0/24'
                            (code, reply) = postconf.inputbox(text=msg, width=len(msg)+4, init=''.join(server_config['list_subnet_skip_login_wads']))
                            server_config['list_subnet_skip_login_wads'] = reply.split(',')
                        else:
                            server_config['list_subnet_skip_login_wads'] = []
            else:
                server_config['wads_enable'] = False

        #wua activation
        if not quiet:
            if not options.secured_mode:
                if postconf.yesno("Do you want to activate WUA ?") == postconf.DIALOG_OK:
                    server_config['waptwua_enable'] = True
                else:
                    server_config['waptwua_enable'] = False
            else:
                server_config['waptwua_enable'] = False

        # Guess fqdn using socket
        fqdn = guess_fqdn()

        clients_signing_certificate = server_config.get('clients_signing_certificate')
        clients_signing_key = server_config.get('clients_signing_key')

        if not clients_signing_certificate or not clients_signing_key:
            clients_signing_certificate = os.path.join(wapt_root_dir, 'conf', 'ca-%s.crt' % fqdn)
            clients_signing_key = os.path.join(wapt_root_dir, 'conf', 'ca-%s.pem' % fqdn)

            server_config['clients_signing_certificate'] = clients_signing_certificate
            server_config['clients_signing_key'] = clients_signing_key

        if clients_signing_certificate is not None and clients_signing_key is not None and not os.path.isfile(clients_signing_certificate):
            print('Create a certificate and key for clients certificate signing')

            key = SSLPrivateKey(clients_signing_key)
            if not os.path.isfile(clients_signing_key):
                print('Create SSL RSA Key %s' % clients_signing_key)
                key.create()
                key.save_as_pem()

            crt = key.build_sign_certificate(cn=fqdn, is_code_signing=False, is_ca=True)
            print('Create X509 cert %s' % clients_signing_certificate)
            crt.save_as_pem(clients_signing_certificate)

        if not server_config['ssl_additional_crls']:
            server_config['ssl_additional_crls'] = server_config['known_certificates_folder']
        if not server_config['clients_signing_crl']:
            # somewhere reachable using http
            server_config['clients_signing_crl'] = os.path.join(server_config['known_certificates_folder'],'ca-%s.crl' % fqdn)
        if not server_config['clients_signing_crl_url']:
            server_config['clients_signing_crl_url'] = 'http://%s/wapt/ssl/ca-%s.crl' % (fqdn,fqdn)

        if not os.path.isdir(server_config['known_certificates_folder']):
            mkdirs(server_config['known_certificates_folder'])
        if not os.path.isdir(server_config['ssl_additional_crls']):
            mkdirs(server_config['ssl_additional_crls'])

        clients_signing_crl = server_config.get('clients_signing_crl')
        if clients_signing_certificate is not None and clients_signing_key is not None and clients_signing_crl is not None and not os.path.isfile(clients_signing_crl):
            print('Create a CRL for clients certificate signing')
            key = SSLPrivateKey(clients_signing_key)
            crt = SSLCertificate(clients_signing_certificate)
            crl = SSLCRL(clients_signing_crl, cacert=crt, cakey=key)
            crl.revoke_cert()
            crl.save_as_pem()

        waptserver.config.write_config_file(cfgfile=options.configfile, server_config=server_config, non_default_values_only=True)

        # append a [uwsgi] section if needed
        if server_config['use_uwsgi']:
            append_uwsgi_config(cfgfile=options.configfile)

        server_config = waptserver.config.load_config(cfgfile=options.configfile)

        print('[*] Protecting WAPT config directory')
        run("/bin/chmod 750 %s" % (os.path.abspath(os.path.join(options.configfile, os.pardir))))
        run("/bin/chown wapt:%s %s" % (NGINX_ACCOUNT, os.path.abspath(os.path.join(options.configfile, os.pardir))))
        run("/bin/chmod 700 %s" % (os.path.join(os.path.abspath(os.path.join(options.configfile, os.pardir)), '*')))
        run("/bin/chown wapt:root %s" % (os.path.join(os.path.abspath(os.path.join(options.configfile, os.pardir)), '*')))
        run("/bin/chmod 750 %s" % (os.path.join(os.path.abspath(os.path.join(options.configfile, os.pardir)), '*.crt')))
        run("/bin/chown wapt:%s %s" % (NGINX_ACCOUNT, os.path.join(os.path.abspath(os.path.join(options.configfile, os.pardir)), '*.crt')))

        print('[*] Update WAPT repository')
        repo = WaptLocalRepo(wapt_folder)
        repo.update_packages_index()

        if options.secured_mode:
            update_conf_for_secure_mode(server_config, wapt_root_dir = wapt_root_dir)
            server_config = waptserver.config.load_config(cfgfile=options.configfile)

        final_msg = ['[*] Postconfiguration completed.', ]

        # Nginx configuration
        if quiet:
            if nginx:
                try:
                    generate_dhparam(options.dhparam_key_size)
                    nginx_cleanup()
                    make_nginx_config('/opt/wapt/waptserver', fqdn, options.force_https, server_config=server_config, server_names=server_names)
                    update_all_crls(server_config)
                    enable_nginx()
                    restart_nginx()
                    setup_firewall()
                except subprocess.CalledProcessError as cpe:
                    final_msg += [
                        'Error while trying to configure Nginx!',
                        'errno = %s, output: %s' % (cpe.returncode, cpe.output)
                    ]
                except Exception:
                    import traceback
                    final_msg += [
                        'Error while trying to configure Nginx!',
                        traceback.format_exc()
                    ]
        else:
            reply = postconf.yesno("Do you want to configure nginx?")
            if reply == postconf.DIALOG_OK:
                try:
                    msg = 'FQDN for the WAPT server (eg. wapt.acme.com)'
                    (code, reply) = postconf.inputbox(text=msg, width=len(msg)+4, init=fqdn)
                    if code != postconf.DIALOG_OK:
                        exit(1)
                    else:
                        fqdn = reply

                    generate_dhparam(options.dhparam_key_size)
                    nginx_cleanup()

                    if server_config['use_kerberos']:
                        if type_debian():
                            if not check_if_deb_installed('libnginx-mod-http-auth-spnego'):
                                print('[*] Nginx - Missing dependency libnginx-mod-http-auth-spnego, please install first before configuring kerberos')
                                sys.exit(1)

                        if not os.path.isfile(keytab_file):
                            generate_keytab()

                    make_nginx_config('/opt/wapt/waptserver', fqdn, options.force_https, server_config=server_config, server_names=server_names)
                    update_all_crls(server_config)
                    postconf.msgbox("The Nginx config is done. We need to restart Nginx?")
                    print(enable_nginx())
                    print(restart_nginx())
                    setup_firewall()

                except subprocess.CalledProcessError as cpe:
                    final_msg += [
                        'Error while trying to configure Nginx!',
                        'errno = %s  output: %s' % (cpe.returncode,cpe.output)
                    ]
                except Exception:
                    import traceback
                    final_msg += [
                        'Error while trying to configure Nginx!',
                        traceback.format_exc()
                    ]

        # known certificates
        ssl_dir = server_config['known_certificates_folder']
        if not os.path.isdir(ssl_dir):
            # move existing ssl dir in wapt repo to parent dir (default location)
            if os.path.isdir(os.path.join(server_config['wapt_folder'], 'ssl')):
                shutil.move(os.path.join(server_config['wapt_folder'], 'ssl'), ssl_dir)
            else:
                os.makedirs(ssl_dir)

        # Rename drivers conf for new version
        wads_folder = server_config['wads_folder']
        if os.path.isdir(wads_folder):
            for driver_bundle_filename in glob.glob(r"%s/drivers/conf/*" % wads_folder):
                if not os.path.isfile(driver_bundle_filename):
                    continue
                with open(driver_bundle_filename) as f:
                    data = json.loads(f.read())
                hash_bundle_filename = hashlib.sha256(data['name'].encode('utf-8')).hexdigest()
                if driver_bundle_filename.rsplit('/', 1)[1] != hash_bundle_filename:
                    dstfilename = os.path.join(driver_bundle_filename.rsplit('/', 1)[0], hash_bundle_filename )
                    print('Move %s --> %s' % (driver_bundle_filename, dstfilename))
                    shutil.move(driver_bundle_filename, dstfilename)


        if server_config['homepage_enable'] :
            final_msg.append('Please connect to https://' + fqdn + '/ to access the server.')

        WAPT_UID = good_pwd.getpwnam('wapt').pw_uid

        # CHOWN of waptservertasks.sqlite it seems to be created before
        location_waptservertasks = os.path.join(wapt_root_dir, 'db', 'waptservertasks.sqlite')
        if os.path.isfile(location_waptservertasks):
            os.chown(location_waptservertasks, WAPT_UID, os.stat(location_waptservertasks).st_gid)

        # Create empty sync.json and rules.json file for all installations
        sync_json = os.path.join(os.path.abspath(os.path.join(wapt_folder, os.pardir)), 'sync.json')
        rules_json = os.path.join(os.path.abspath(os.path.join(wapt_folder, os.pardir)), 'rules.json')
        licences_json = os.path.join(os.path.abspath(os.path.join(wapt_folder, os.pardir)), 'licences.json')
        diff_rules_dir = wapt_folder+'-diff-repos'
        wapt_host = wapt_folder+'-host'
        waptwua = wapt_folder+'wua'
        wuafolder = server_config['waptwua_folder']

        paths_to_modify = [(licences_json, True), (sync_json, True), (rules_json, True), (wapt_folder, False), (wuafolder, False), (diff_rules_dir, False), (ssl_dir, False), (waptwua, False), (wapt_host, False)]

        if server_config['clients_signing_crl'] is not None:
            paths_to_modify.append((server_config['clients_signing_crl'], True))
        if server_config['ssl_client_crls'] is not None:
            paths_to_modify.append((server_config['ssl_client_crls'], True))

        for apath, isfile in paths_to_modify:
            if os.path.isdir(apath):
                os.chown(apath, WAPT_UID, NGINX_GID)
                os.chmod(apath, 0o750)
                for root, dirs, files in os.walk(apath):
                    for d in dirs:
                        full_path = os.path.join(root, d)
                        os.chown(full_path, WAPT_UID, NGINX_GID)
                        os.chmod(full_path, 0o750)
                    for f in files:
                        full_path = os.path.join(root, f)
                        os.chown(full_path, WAPT_UID, NGINX_GID)
                        os.chmod(full_path, 0o640)
            elif not(isfile):
                mkdirs(apath)
                os.chmod(apath, 0o750)
                os.chown(apath, WAPT_UID, NGINX_GID)
            else:
                if not(os.path.isfile(apath)):
                    with open(apath, 'w') as f:
                        if os.path.basename(apath) in ('licences.json','rules.json'):
                            f.write('[]')
                        elif os.path.basename(apath) in ('sync.json'):
                            f.write('{}')
                    os.chmod(apath, 0o640)
                    if type_redhat():
                        run('restorecon -v %s' % (apath))
                os.chown(apath, WAPT_UID, NGINX_GID)

        if not quiet:
            postconf.msgbox("Press ok to start waptserver and wapttasks daemons")
        out = enable_waptserver()
        if server_config['use_uwsgi']:
            out += run('systemctl enable waptserver-uwsgi')
        else:
            if service_is_enabled('waptserver-uwsgi'):
                out += run('systemctl disable waptserver-uwsgi')

        out += start_waptserver()
        if not quiet:
            print(out)

        #run for populate database
        run("/opt/wapt/wapt-scanpackages.sh -b %s" % server_config['wapt_folder']).strip()

        # Final message
        if not quiet:
            width = 4 + max(10, len(max(final_msg, key=len)))
            height = 2 + max(20, len(final_msg))
            postconf.msgbox('\n'.join(final_msg), height=height, width=width)
        else:
            if wapt_password:
                final_msg.append('[*] WAPT admin password : %s\n' % wapt_password)
            for line in final_msg:
                print(line)

    finally:
        # Configuration summary
        print("""\
Waptserver
    Version: %s

""" % ( get_wapt_full_version(),))

        ssl_dir = os.path.join('/opt/wapt/waptserver', 'ssl')
        wapt_ssl_cert_file = os.path.join(ssl_dir, 'cert.pem')
        if os.path.isfile(wapt_ssl_cert_file):
            crt = SSLCertificate(wapt_ssl_cert_file)
            print("""\
Waptserver certificate
 Subject:   %s
    Issuer :   %s
    Altnames:  %s
    SHA1:      %s
    SHA256:    %s

"""
 % ( crt.subject, crt.issuer_dn, ', '.join(str(s) for s in crt.subject_alt_names), crt.fingerprint_sha1, crt.fingerprint))

            waptsetup_path = os.path.join(server_config['wapt_folder'],'waptagent','waptsetup.exe')
            if os.path.isfile(waptsetup_path):
                fqdn = guess_fqdn()
                hostnames = [str(n) for n in crt.subject_alt_names if fqdn in str(n)]
                if hostnames:
                    waptsetup_url = 'https://%s/wapt/waptagent/waptsetup.exe' % hostnames[0]
                    waptsetup_sha256 = sha256_for_file(waptsetup_path)
                    print("""\
Waptsetup agent installer:
    Version available on server: %s
    You can download "waptsetup" Windows agent installer at URL: %s
    Check the Code signing signature and the hash of the installer: Sha256: %s

""" % (get_wapt_exe_version(waptsetup_path), waptsetup_url, waptsetup_sha256))

            else:
                print("""\
Waptsetup Windows agent installer is not present on your server.\
You can download it from Tranquil IT repositories.
""")



if __name__ == "__main__":
    if not type_debian() and not type_redhat():
        print("Unsupported distribution")
        sys.exit(1)
    main()

