#!/usr/bin/env python3
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2023  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##
import os
import time
import datetime
import json
import ujson
import logging
import traceback
import glob
import hashlib
import tempfile
import shutil
import jinja2
import ipaddress
from flask import request, Blueprint, session, redirect
from werkzeug.utils import secure_filename
from waptserver.model import wapt_db, wapt_db_connect, IsoConfig, Hosts, DeployConfig, HostOSDeploy, WaptDB
from waptserver.decorators import requires_auth, require_wapt_db, requires_enterprise, wads_enable, authenticate, forbidden
from waptserver.common import make_response, make_response_from_exception, PostDataStream, get_server_uuid, get_secured_token_generator,get_secondary_repo
from waptserver.common import licences_list # use as global below
from waptserver.enterprise import prepare_request_data, update_or_insert
from waptserver.auth import check_auth, get_user_acls, has_any_right
from waptserver.app import app
from waptserver import repositories
from waptserver.utils import get_forwarded_url, get_hash256_exe
from waptutils import CustomZipFile

try:
    # i18n
    from flask_babel import Babel
    from flask_babel import gettext
except ImportError:
    gettext = (lambda s: s)
    Babel = None

_ = gettext

wads = Blueprint('wads', __name__)

logger = logging.getLogger('waptserver')

@wads.route('/api/v3/get_iso_config', methods=['GET'])
@requires_auth(['admin', 'wads_view', 'wads_host_deploy', 'wads_admin'])
@requires_enterprise
@require_wapt_db
def get_iso_config():
    def check_iso_in_list(iso_name, iso_config_list):
        counter = 0
        for iso in iso_config_list:
            if iso['hash'] + '.iso' == iso_name:
                iso['file_exists'] = 'true'
                counter+=1
        return counter

    start_time = time.time()
    wapt_db_connect()
    try:
        data = IsoConfig.select()
        data = list(data.dicts())

        iso_folder = os.path.join(app.conf['wads_folder'],'iso')
        for iso_name in glob.glob(r"%s/**" % iso_folder):
            if os.path.isfile(iso_name):
                if check_iso_in_list(os.path.basename(iso_name), data) == 0:
                    new_iso = dict(
                        file_exists = 'false',
                        name = os.path.basename(iso_name),
                        hash = os.path.basename(iso_name)[0:-4],
                        created_on = datetime.datetime.fromtimestamp(os.path.getmtime(iso_name))
                    )
                    data.append(new_iso)

        return make_response(data, success=True, request_time=time.time() - start_time)
    except Exception as e:
        return make_response_from_exception(e)

@wads.route('/api/v3/delete_iso', methods=['POST'])
@requires_auth(['admin'])
@require_wapt_db
def delete_iso():
    try:
        iso_to_delete = ujson.loads(request.data)
        if not iso_to_delete:
            raise Exception("ISO : No data supplied")

        iso_files_folder = os.path.join(app.conf['wads_folder'],'iso')
        if not os.path.isdir(iso_files_folder):
                return make_response(success=False, msg='The Directory %s does not exist.' % (os.path.join(app.conf['wads_folder'],'iso')));

        for filepath_iso in iso_to_delete:
            filepath_iso = filepath_iso + '.iso'
            file_to_delete = os.path.join(iso_files_folder,secure_filename(filepath_iso))
            logger.info('Deleting iso %s' % (file_to_delete))
            if os.path.isfile(file_to_delete):
                os.unlink(file_to_delete)

        global licences_list
        if licences_list.is_enterprise and Hosts.select().where(Hosts.wapt_status.contains({'is_remote_repo':True})).count():
            repositories.update_file_tree_of_files(username=session.get('user'))

        return make_response(success= not os.path.isfile(file_to_delete))

    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('delete iso failed %s' % (repr(e)))
        return make_response_from_exception(e)

@wads.route('/api/v3/delete_iso_config', methods=['GET','POST'])
@requires_auth(['admin'])
@requires_enterprise
@require_wapt_db
def delete_iso_config():
    start_time = time.time()
    prepare_request_data(request)
    try:
        iso_to_delete = ujson.loads(request.data)
        if not iso_to_delete:
            raise Exception("Iso : No data supplied")
    except Exception as e:
        return make_response_from_exception(e)

    with wapt_db.atomic() as db:
        try:
            cnt = IsoConfig.delete().where(IsoConfig.id.in_(iso_to_delete)).execute()
            return make_response(cnt, success=True, request_time=time.time() - start_time)
        except Exception as e:
            db.rollback()
            return make_response_from_exception(e)

@wads.route('/api/v3/get_deploy_config', methods=['GET'])
@requires_auth(['admin', 'wads_view', 'wads_host_deploy', 'wads_admin'])
@requires_enterprise
@require_wapt_db
def get_deploy_config():
    start_time = time.time()
    wapt_db_connect()
    try:
        data = DeployConfig.select(DeployConfig, IsoConfig.name.alias('isoconfig_name')).join(IsoConfig, 'LEFT JOIN', on=(DeployConfig.iso_config_name == IsoConfig.name))
        data = list(data.dicts())
        return make_response(data, success=True, request_time=time.time() - start_time)
    except Exception as e:
        return make_response_from_exception(e)

@wads.route('/api/v3/save_deploy_config', methods=['POST'])
@requires_auth(['admin', 'wads_admin'])
@requires_enterprise
@require_wapt_db
def save_deploy_config():
    start_time = time.time()

    prepare_request_data(request)
    delta_recs = ujson.loads(request.data)

    try:
        result = update_or_insert(DeployConfig, delta_recs,session.get('user'))
        return make_response(result=result, success=True, request_time=time.time() - start_time)
    except Exception as e:
        return make_response_from_exception(e)

@wads.route('/api/v3/delete_deploy_config', methods=['GET','POST'])
@requires_auth(['admin', 'wads_admin'])
@requires_enterprise
@require_wapt_db
def delete_deploy_config():
    start_time = time.time()
    prepare_request_data(request)
    try:
        config_to_delete = ujson.loads(request.data)
        if not config_to_delete:
            raise Exception("Deploy Config : No data supplied")
    except Exception as e:
        return make_response_from_exception(e)

    with wapt_db.atomic() as db:
        try:
            cnt = DeployConfig.delete().where(DeployConfig.id.in_(config_to_delete)).execute()
            return make_response(cnt, success=True, request_time=time.time() - start_time)
        except Exception as e:
            db.rollback()
            return make_response_from_exception(e)

@wads.route('/api/v3/get_driver_bundle_names', methods=['GET'])
@requires_auth(['admin', 'wads_view', 'wads_host_deploy', 'wads_admin'])
@requires_enterprise
@require_wapt_db
def get_driver_bundle_names():
    name_list = []

    driver_bundle_folder = os.path.join(app.conf['wads_folder'],'drivers','conf')
    if not os.path.isdir(driver_bundle_folder):
        pass
    else:
        for driver_bundle_filename in glob.glob(r"%s/**" % driver_bundle_folder):
            if not os.path.isfile(driver_bundle_filename):
                continue
            with open(driver_bundle_filename) as f:
                j = json.loads(f.read())
                j['id'] = hashlib.sha256(os.path.basename(j['name']).encode('utf-8')).hexdigest()
                name_list.append(j)

    return make_response(result=name_list, success=True)

@wads.route('/api/v3/upload_deploy_files', methods=['POST'])
@requires_auth(['admin', 'wads_admin'])
@require_wapt_db
def upload_deploy_files():
    try:
        starttime = time.time()

        upload_folder = os.path.join(app.conf['wads_folder'])
        folder_list = request.args.get('folder', '')
        if folder_list != '':
            folders = folder_list.split(',')
            for folder in folders:
                upload_folder = os.path.join(upload_folder, folder)

        if not os.path.isdir(upload_folder):
            os.makedirs(upload_folder)

        filename = request.args.get('filename')
        overwrite = bool(int(request.args.get('overwrite', 0)))
        hash_file = request.args.get('hash', '')
        check_need_upload = bool(int(request.args.get('check_need_upload', 0)))

        target = os.path.join(upload_folder,secure_filename(filename))

        upload = False

        if os.path.isfile(target):
            if overwrite:
                if hash_file == '':
                    if not check_need_upload:
                        os.unlink(target)
                    upload = True
                elif get_hash256_exe(target) != hash_file:
                    if not check_need_upload:
                        os.unlink(target)
                    upload = True
            else:
                if not check_need_upload:
                    raise Exception("File already exists")
        else:
            upload = True

        if upload and (not check_need_upload):
            # streamed upload
            logger.info('Start streaming down file to %s' % target)
            filestream = PostDataStream(request.stream,calc_sha256 = (hash_file != ''))
            try:
                filestream.save(target)
                if hash_file != '':
                    if filestream.sha256digest() != hash_file:
                        logger.info('Deleting file %s due to wrong hash' % target)
                        raise Exception("wrong hash")
            except Exception as e:
                if os.path.isfile(target):
                    logger.info('Deleting file %s due to %s' % (target, e))
                    os.unlink(target)
                raise

        if check_need_upload:
            if upload and os.path.isfile(target):
                success_state = False
                response_msg = _('NEED UPLOAD {} present. Size {}').format(target,os.stat(target).st_size)
            elif upload and not os.path.isfile(target):
                success_state = False
                response_msg = _('NEED UPLOAD {} not present.').format(target)
            elif (not upload) and os.path.isfile(target):
                success_state = True
                response_msg = _('NO NEED UPLOAD {} present. Size {}').format(target,os.stat(target).st_size)
            else:
                success_state = False
                response_msg = _('NEED UPLOAD NO FILE THERE {}').format(target)
        else:
            success_state = os.path.isfile(target)
            if success_state:
                response_msg = _('{} present. Size {}').format(target,os.stat(target).st_size)
            else:
                response_msg = _('{} not present.').format(target)

            global licences_list
            if licences_list.is_enterprise and Hosts.select().where(Hosts.wapt_status.contains({'is_remote_repo':True})).count():
                repositories.update_file_tree_of_files(username=session.get('user'))

        return make_response(success=success_state,
                             result=target,
                             msg=response_msg,
                             request_time=time.time() - starttime)

    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('upload deploy files (%s) failed %s' % (filename ,repr(e)))
        return make_response_from_exception(e)

@wads.route('/api/v3/delete_driver_bundles', methods=['POST'])
@requires_auth(['admin', 'wads_admin'])
@require_wapt_db
def delete_driver_bundles():
    try:
        bundles_to_delete = ujson.loads(request.data)
        if not bundles_to_delete:
            raise Exception("No list of driver bundles names supplied")

        driver_bundle_folder = os.path.join(app.conf['wads_folder'],'drivers', 'conf')
        if not os.path.isdir(driver_bundle_folder):
            return make_response(success=False, msg='The Directory %s does not exist.' % (driver_bundle_folder,));

        for bundle_name in bundles_to_delete:
            bundle_filename = os.path.join(driver_bundle_folder,secure_filename(hashlib.sha256(bundle_name.encode('utf-8')).hexdigest()))
            logger.info('Deleting bundle %s' % (hashlib.sha256(bundle_name.encode('utf-8')).hexdigest()))
            if os.path.isfile(bundle_filename):
                os.unlink(bundle_filename)

        global licences_list
        if licences_list.is_enterprise and Hosts.select().where(Hosts.wapt_status.contains({'is_remote_repo':True})).count():
            repositories.update_file_tree_of_files(username=session.get('user'))
        return make_response()

    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('delete drivers bundle failed %s' % (repr(e)))
        return make_response_from_exception(e)

@wads.route('/api/v3/delete_unused_driver_files', methods=['DELETE'])
@requires_auth(['admin', 'wads_admin'])
@require_wapt_db
def delete_unused_driver_files():
    filehashes_refcount = {}
    driver_files_folder = os.path.join(app.conf['wads_folder'],'drivers','files')
    driver_bundle_folder = os.path.join(app.conf['wads_folder'],'drivers','conf')
    if not os.path.isdir(driver_files_folder):
        pass
    else:
        for bundle_filename in glob.glob(r"%s/**" % driver_bundle_folder):
            if not os.path.isfile(bundle_filename):
                continue
            with open(bundle_filename) as f:
                driver_bundle = json.load(f)
            """
                {"name":"test",
                 "data":
                      {"filehash1": ["filename1","filename2",..],
                       "filehash2": ["filename4",'filename5",..],
                      },
                 "signer": "xxxx",
                 "signature": "llll",
                 "signature_date": :"2022-02-14T10:50:49",
                 "signed_attributes": ["signer","signature_date","signed_attributes","data","name"]

                }
            """
            for (filehash,filenames) in driver_bundle['data'].items():
                if not filehash in filehashes_refcount:
                    filehashes_refcount[filehash] = len(filenames)
                else:
                    filehashes_refcount[filehash] += len(filenames)

        for filename in glob.glob(r"%s/**" % driver_files_folder):
            if not os.path.isfile(filename):
                continue
            if not filehashes_refcount.get(os.path.basename(filename)):
                os.unlink(filename)

    global licences_list
    if licences_list.is_enterprise and Hosts.select().where(Hosts.wapt_status.contains({'is_remote_repo':True})).count():
        repositories.update_file_tree_of_files(username=session.get('user'))

    return make_response(success=True)

@wads.route('/api/v3/get_driver_filehashes', methods=['GET'])
@requires_auth(['admin', 'wads_view', 'wads_host_deploy', 'wads_admin'])
def get_driver_filehashes():
    file_list = []

    driver_files_folder = os.path.join(app.conf['wads_folder'],'drivers','files')
    if not os.path.isdir(driver_files_folder):
        pass
    else:
        for file in glob.glob(r"%s/**" % driver_files_folder):
            if not os.path.isfile(file):
                continue
            file_list.append(os.path.basename(file))

    return make_response(result=file_list, success=True)

@wads.route('/api/v3/get_host_deploy', methods=['GET'])
@requires_auth(['admin', 'wads_admin', 'wads_host_deploy', 'wads_view'])
@requires_enterprise
@require_wapt_db
def get_host_deploy():
    start_time = time.time()
    wapt_db_connect()
    try:
        data = list(HostOSDeploy.select().dicts())
        return make_response(data, success=True, request_time=time.time() - start_time)
    except Exception as e:
        return make_response_from_exception(e)

@wads.route('/api/v3/save_host_deploy', methods=['POST'])
@requires_auth(['admin', 'wads_host_deploy', 'wads_admin'])
@requires_enterprise
@require_wapt_db
def save_host_deploy():
    start_time = time.time()

    prepare_request_data(request)
    hosts = ujson.loads(request.data)

    try:
        result = update_or_insert(HostOSDeploy, hosts, session.get('user'))
        return make_response(result=result, success=True, request_time=time.time() - start_time)
    except Exception as e:
        return make_response_from_exception(e)

@wads.route('/api/v3/delete_host_deploy', methods=['GET','POST'])
@requires_auth(['admin', 'wads_host_deploy', 'wads_admin'])
@requires_enterprise
@require_wapt_db
def delete_host_deploy():
    start_time = time.time()
    prepare_request_data(request)
    try:
        host_to_delete = ujson.loads(request.data)
        if not host_to_delete:
            raise Exception("Host to Deploy : No data supplied")
    except Exception as e:
        return make_response_from_exception(e)

    with wapt_db.atomic() as db:
        try:
            cnt = HostOSDeploy.delete().where(HostOSDeploy.id.in_(host_to_delete)).execute()
            return make_response(cnt, success=True, request_time=time.time() - start_time)
        except Exception as e:
            db.rollback()
            return make_response_from_exception(e)

@wads.route('/api/v3/import_host_from_inventory', methods=['GET'])
@requires_auth(['admin', 'wads_host_deploy', 'wads_admin'])
@requires_enterprise
@require_wapt_db
def import_host_from_inventory():
    start_time = time.time()
    wapt_db_connect()
    try:
        list_mac = []
        data = HostOSDeploy.select(HostOSDeploy.mac_addresses).dicts()
        for entry in data:
            list_mac.extend([mac.lower() for mac in entry['mac_addresses'].split(',')])

        data = list(Hosts.select().dicts())
        dict_hostname_mac = {}
        for entry in data:
            try:
                doublon = False
                new_mac_list = []
                if entry['wmi'] :
                    for mac in entry['wmi']['Win32_NetworkAdapter'] :
                        if 'wifi' in str(mac).lower() or "bluetooth" in str(mac).lower() or 'wireless' in str(mac).lower() or 'wi-fi' in str(mac).lower() or 'sans fil' in str(mac).lower() :
                            continue
                        if not mac.get('PhysicalAdapter'):
                            continue
                        if mac['PNPDeviceID'].startswith('ROOT\\'):
                            continue
                        maclower = mac['MACAddress'].lower()
                        if not maclower in list_mac:
                            new_mac_list.append(maclower)
                            list_mac.append(maclower)
                        else:
                            doublon =True
                            break
                else:
                    for mac in entry['mac_addresses']:
                        maclower = mac.lower()
                        if not maclower in list_mac:
                            new_mac_list.append(maclower)
                            list_mac.append(maclower)
                        else:
                            doublon =True
                            break
                if doublon:
                    continue
                if not new_mac_list:
                    continue
                dict_hostname_mac[entry['computer_name']]=new_mac_list
                serialnr = ''
                manufacturer= ''
                product = ''
                uuid = ''
                try:
                    uuid = entry['dmi']['System_Information']['UUID']
                    if not 'Base_Board_Information' in entry['dmi']:
                         serialnr = entry['serialnr']
                         manufacturer= entry['manufacturer']
                         product = entry['host_info']['system_productname']
                    else:
                         serialnr = entry['dmi']['Base_Board_Information']['Serial_Number']
                         manufacturer = entry['dmi']['Base_Board_Information']['Manufacturer']
                         product = entry['dmi']['Base_Board_Information']['Product_Name']
                except:
                    pass

                host_data = dict(
                        mac_addresses=','.join(new_mac_list),
                        hostname=entry['computer_name'].split('.')[0],
                        uuid=uuid,
                        manufacturer=manufacturer,
                        serial_number=serialnr,
                        waiting_to_deploy=False,
                        djoin="",
                        status="imported from inventory",
                        product = product,
                        created_on = datetime.datetime.utcnow(),
                        created_by = session.get('user')
                )

                HostOSDeploy.insert(**host_data).execute()
            except:
                pass
        data  = dict_hostname_mac
        return make_response(data, success=True, request_time=time.time() - start_time)
    except Exception as e:
        return make_response_from_exception(e)

@wads.route('/api/v3/import_host_from_json', methods=['POST'])
@requires_auth(['admin', 'wads_host_deploy', 'wads_admin'])
@requires_enterprise
@require_wapt_db
def import_host_from_json():
    start_time = time.time()
    prepare_request_data(request)
    try:
        hosts_received = ujson.loads(request.data)
        if not hosts_received:
            raise Exception("Hosts : No data to import")
    except Exception as e:
        return make_response_from_exception(e)
    wapt_db_connect()
    try:
        existing_macs = []
        data = list(HostOSDeploy.select(HostOSDeploy.mac_addresses).tuples())
        for entry in data:
            existing_macs.extend([mac.lower() for mac in entry[0].split(',')])
        new_hosts = []
        for new_host in hosts_received:
            if len(new_host['hostname']) > 15:
                continue
            try:
                new_host['mac_addresses'] = new_host['mac_addresses'].replace('-',':').lower()
                new_host_macs = new_host['mac_addresses'].split(',')
                for new_host_mac in new_host_macs:
                    new_host_mac = new_host_mac.lower()
                    if not new_host_mac in existing_macs:
                        make_insert = True
                        if len(new_host_macs) > 1:
                            for the_mac in new_host_macs:
                                the_mac = the_mac.lower()
                                if the_mac in existing_macs:
                                    make_insert = False
                                    break
                        if make_insert:
                            host_data = dict(
                                mac_addresses = new_host['mac_addresses'].lower(),
                                status = "imported from json",
                                waiting_to_deploy = False,
                                uuid = new_host['uuid'],
                                hostname = new_host['hostname'],
                                deploy_config_name = new_host['deploy_config_name'],
                                product_key = new_host['product_key'],
                                driver_bundle = new_host['driver_bundle'],
                                disk_format = new_host['disk_format'],
                                djoin = new_host['djoin'],
                                manufacturer = new_host['manufacturer'],
                                product = new_host['product'],
                                serial_number = new_host['serial_number'],
                                asset = new_host['asset'],
                                ei_cfg = new_host['ei_cfg'],
                                pid_txt = new_host['pid_txt'],
                                diskpart_list_disk = new_host['diskpart_list_disk'],
                                diskpart_list_vol = new_host['diskpart_list_vol'],
                                ip_address = new_host['ip_address'],
                                created_on = datetime.datetime.utcnow(),
                                created_by = session.get('user')
                            )
                            HostOSDeploy.insert(**host_data).execute()
                            existing_macs.extend(new_host_macs)
                            new_hosts.append(host_data)
                            break

            except Exception as e:
                print('erreur: %s' % e)

        data  = new_hosts
        return make_response(data, success=True, request_time=time.time() - start_time)
    except Exception as e:
        return make_response_from_exception(e)

@wads.route('/api/v3/upload_winpe', methods=['POST'])
@requires_auth(['admin','wads_admin'])
@require_wapt_db
def upload_winpe():
    try:
        starttime = time.time()

        winpe_file_folder = os.path.join(app.conf['wads_folder'],'winpe', request.args.get('arch','x64') ,'media')
        if not os.path.isdir(winpe_file_folder):
            os.makedirs(winpe_file_folder)

        tmp_target = tempfile.mktemp(dir=winpe_file_folder)

         # streamed upload
        logger.info('Start streaming down file to %s' % tmp_target)
        filestream = PostDataStream(request.stream)
        filestream.save(tmp_target)

        if request.args.get('filename'):
            target = os.path.join(winpe_file_folder,secure_filename(request.args.get('filename')))

            if os.path.isfile(target):
                os.unlink(target)
            logger.info('Renaming iso %s into %s' % (tmp_target, target))
            shutil.move(tmp_target, target)
        else:
            target = tmp_target

        with CustomZipFile(target, 'r') as target_zip:
            target_zip.extractall(winpe_file_folder)

        target_size = os.stat(target).st_size
        os.remove(target)

        #####################   For use without ipxe
        if app.conf['copy_winpe_x64_in_tftp_folder']:
            if request.args.get('arch','x64') == 'x64':
                pxe_file_folder = os.path.join(app.conf['wads_folder'],'pxe')
                list_folder_copy = ['Boot','EFI','sources']
                for fpxe in list_folder_copy:
                    dstfolder=os.path.join(pxe_file_folder,fpxe)
                    if os.path.isdir(dstfolder):
                        shutil.rmtree(dstfolder)
                    shutil.copytree(os.path.join(winpe_file_folder,fpxe),dstfolder)
                shutil.copyfile(os.path.join(pxe_file_folder,"EFI","Microsoft","Boot","BCD"),os.path.join(pxe_file_folder,"Boot","BCD"))
        ################################

        spenttime = time.time() - starttime
        global licences_list
        if licences_list.is_enterprise and Hosts.select().where(Hosts.wapt_status.contains({'is_remote_repo':True})).count():
            repositories.update_file_tree_of_files(username=session.get('user'))

        return make_response(success=os.path.isfile(target),
                             result=target,
                             msg=_('{} uploaded. Size {}').format(target, target_size),
                             request_time=spenttime)
    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('upload_wwinpe failed %s' % (repr(e)))
        return make_response_from_exception(e)

@wads.route('/api/v3/check_winpe_exist', methods=['GET'])
@requires_auth(['admin','view','wads_view', 'wads_host_deploy', 'wads_admin'])
def check_winpe_exist():
    exist = False
    winpe_file_folder = os.path.join(app.conf['wads_folder'],'winpe', 'x64' ,'media', 'Boot')
    if os.path.isdir(winpe_file_folder):
        if os.path.isfile(os.path.join(winpe_file_folder,secure_filename('boot.sdi'))):
            exist = True
    return make_response(success=exist)

@wads.route('/2PXE/Boot', methods=['POST'])
@wads.route('/api/v3/baseipxe', methods=['GET','POST'])
def get_base_ipxe():
    if not app.conf['wads_enable']:
        return "#!ipxe\nset esc:hex 1b\nset cls ${esc:string}[2J\necho ${cls}\necho Wads is not activated on the wapt server ... continue\nsleep 10\nexit"
    uefi = request.args.get('uefi', "true")

    keymap = request.args.get('keymap')
    keymapcommand=""
    if keymap:
        keymapcommand = "set keymap %s" % keymap

    ignore_ipxescript = request.args.get('ignore_ipxescript', "false")
    return """#!ipxe
%s
chain %sapi/v3/get_host_ipxe?mac_address=${net0/mac}&ip_address=${net0/ip}&uefi=%s&uuid=${uuid}&ignore_ipxescript=%s&manufacturer=${manufacturer}&product=${product}&serial_number=${serial}&asset=${asset}""" % (keymapcommand,get_forwarded_url(request),uefi,ignore_ipxescript,)

@wads.route('/api/v3/get_host_ipxe', methods=['GET'])
@wads_enable
@require_wapt_db
def get_host_ipxe():
    # GET MAC ADDRESS
    already_pass = request.args.get('already_pass', "false")
    mac_address = request.args.get('mac_address', '').upper()
    ip_address = request.args.get('ip_address', '')
    uefi = request.args.get('uefi', "false")

    manufacturer = request.args.get('manufacturer', "")
    product = request.args.get('product', "")
    serial_number = request.args.get('serial_number', "")
    asset = request.args.get('asset', "")

    ignore_ipxescript = request.args.get('ignore_ipxescript', "false")
    localboot = "exit"
    with WaptDB():
        host = (HostOSDeploy
                    .select(HostOSDeploy.waiting_to_deploy,
                            HostOSDeploy.deploy_config_name,
                            HostOSDeploy.hostname)
                    .where(HostOSDeploy.mac_addresses.contains(mac_address))
                    .first()
                )

    urepo = get_secondary_repo(ip_address,get_forwarded_url(request),repo_name='wads')
    if uefi.lower() == 'false':
        installscript64 = r"""chain %swads/winpe/x64/runinstall_legacy.ipxe
goto start""" % urepo
        installscript86 = r"""chain %swads/winpe/x86/runinstall_legacy.ipxe
goto start""" % urepo
    else:
        installscript64 = r"""chain %swads/winpe/x64/runinstall_uefi.ipxe
goto start""" % urepo
        installscript86 = r"""chain %swads/winpe/x86/runinstall_uefi.ipxe
goto start""" % urepo

    required_wads_login = app.conf['login_on_wads'] and (not skip_login_wads(request.headers.get('X-Forwarded-For', None)))
    if required_wads_login :
        urltemp = get_forwarded_url(request).split('://',1)

        url = urltemp[0] + r'://${username:uristring}:${password:uristring}@' + urltemp[1]
    else:
        url = get_forwarded_url(request)


    if not host:
        j2ipxe = app.conf['ipxe_script_jinja_path']
        with open(j2ipxe,'r') as f:
            tm = jinja2.Template(f.read())


            ipxescript = tm.render(localboot=localboot,
            installscript64=installscript64, installscript86=installscript86, login_wads=required_wads_login, forwarded_url=url,manufacturer=manufacturer,product=product,serial_number=serial_number,asset=asset,mac_address=mac_address)
        return ipxescript

    if not host.waiting_to_deploy:
        return "#!ipxe\nset esc:hex 1b\nset cls ${esc:string}[2J\necho ${cls}\necho %s is not waiting for deployment ... continue\nsleep 4\n%s" % (host.hostname,localboot)
    else:
        host_data = HostOSDeploy.select().where(HostOSDeploy.mac_addresses.contains(mac_address)).dicts().first()
        config_data = DeployConfig.select().where(DeployConfig.name == host_data['deploy_config_name']).dicts().first()
        iso_data = IsoConfig.select().where(IsoConfig.name == config_data['iso_config_name']).dicts().first()
        if config_data.get('ipxe_script'):
            if ignore_ipxescript.lower() == 'false':
                data = config_data.get('ipxe_script')

                if app.conf['login_on_wads'] and (not skip_login_wads(request.headers.get('X-Forwarded-For', None))) :
                    auth_result = check_auth(session=session, request=request, methods=['admin', 'ldap', 'passwd','token','session'])
                    if not auth_result:
                        if already_pass == 'false' :
                            return """#!ipxe\nlogin\nchain %sapi/v3/get_host_ipxe?mac_address=${net0/mac}&ip_address=${net0/ip}&uefi=%s&uuid=${uuid}&ignore_ipxescript=%s&manufacturer=${manufacturer}&product=${product}&serial_number=${serial}&asset=${asset}&already_pass=true""" % (url,uefi,ignore_ipxescript,)
                        else:
                            time.sleep(3)
                            return authenticate()
                    if not has_any_right(get_user_acls(username=auth_result['user']),['admin','wads_host_deploy','wads_admin']):
                        return forbidden()
                    token_gen = get_secured_token_generator()
                    token_data = dict(
                        user=auth_result['user'].lower(),
                        auth_method= auth_result['auth_method'],
                        auth_date=auth_result['auth_date'],
                        server_uuid=get_server_uuid(),
                    )
                    t = token_gen.dumps(token_data)
                    url = url.replace("${password:uristring}",t)
                isohash = '{{isohash}}'
                if iso_data:
                    isohash = iso_data.get('hash')
                return data.replace('{{server_url}}',url).replace('{{secondary_repo}}',get_secondary_repo(ip_address,get_forwarded_url(request))).replace('{{hostname}}',host.hostname).replace('{{isohash}}',isohash)

        ## Test if the Host is awaiting deployment

        arch = iso_data.get('architecture')
        if arch == 'x86':
            installscript = installscript86
        else:
            installscript = installscript64
        ipxescript = r"""#!ipxe
%s""" % installscript

        change_host_wads_status(mac_address, 'Init installation boot')
        return ipxescript

def skip_login_wads(ip):
    for subnet in app.conf['list_subnet_skip_login_wads']:
        try:
            ip_network = ipaddress.ip_network(subnet)
        except ValueError:
            continue
        if bool(ipaddress.ip_address(ip) in ip_network):
            return True
    return False

@require_wapt_db
def change_host_wads_status(mac_address='', host_status=''):
    t = time.localtime()
    current_time = time.strftime("%H:%M:%S %D", t)

    host_status = host_status + " " + current_time

    with WaptDB():
        oldstatus = HostOSDeploy.select().where(HostOSDeploy.mac_addresses.contains(mac_address)).dicts().first()['status']
        HostOSDeploy.update(status=host_status + ' | ' + oldstatus).where(HostOSDeploy.mac_addresses.contains(mac_address)).execute()

    logger.info('Status : %s for %s' % (host_status, mac_address))

@wads.route('/api/v3/get_wads_config', methods=['GET', 'POST'])
@wads_enable
@require_wapt_db
def get_wads_config():
    if request.method == 'POST':
        req = ujson.loads(request.data.decode('utf-8'))
    else:
        req = dict(request.args)

    if app.conf['login_on_wads'] and (not skip_login_wads(request.headers.get('X-Forwarded-For', None))) :
        auth_result = check_auth(session=session, request=request, methods=['session','token','admin', 'ldap', 'passwd'])
        if not auth_result:
            time.sleep(3)
            return authenticate()
        if not has_any_right(get_user_acls(username=auth_result['user']),['admin','wads_host_deploy','wads_admin']):
            return forbidden()

    mac_address = req.get('mac_address', '').upper()
    ip_address = req.get('ip_address', '').upper()
    conf_only = req.get("conf_only", False)

    with WaptDB():
        ## Test if the Host is awaiting deployment
        host_data = HostOSDeploy.select().where(HostOSDeploy.mac_addresses.contains(mac_address)).dicts().first()
        if host_data:
            if not host_data.get('waiting_to_deploy'):
                return make_response(result={'register_host': False}, success=True)
            config_data = DeployConfig.select().where(DeployConfig.name == host_data['deploy_config_name']).dicts().first()
            if not config_data:
                return make_response(result=None,success=False,msg='The config %s does not exist in DB' % host_data['deploy_config_name'])
            if not conf_only:
                iso_data = IsoConfig.select().where(IsoConfig.name == config_data['iso_config_name']).dicts().first()

        if not host_data :
            return make_response(result={'register_host': True}, success=True)

        if (not host_data.get('hostname')) or (not host_data.get('deploy_config_name')) or (not host_data.get('waiting_to_deploy')):
            return make_response(result={'register_host': False}, success=True)

        result = {}
        result['host_data'] = host_data
        result['config_data'] = config_data

        if not conf_only:

            hash_json_filename = os.path.join(app.conf['wads_folder'], 'hash.json')
            if os.path.isfile(hash_json_filename):
                with open(hash_json_filename, 'r') as f:
                    hashes_content = json.loads(f.read())
            else:
                hashes_content = None

            result['iso_data'] = iso_data
            result['wapt_data'] = hashes_content

        logger.info('get_wads_config succeed for host with mac address %s: %s' % (mac_address, result))

        # Delete Djoin for security
        #  see https://www.postgresql.org/docs/current/functions-json.html for jsonb || update operator.
        HostOSDeploy.update({
                HostOSDeploy.djoin: "",
                HostOSDeploy.waiting_to_deploy: False
                }
                ).where(HostOSDeploy.id == host_data['id']).execute()

        if conf_only:
            return config_data['xml_conf'].replace('{{server_url}}',get_forwarded_url(request)).replace('{{secondary_repo}}',get_secondary_repo(ip_address,get_forwarded_url(request))).replace('{{hostname}}',host_data['hostname']).replace('{{djoin}}',str(host_data['djoin']))

    return make_response(result=result)

@wads.route('/api/v3/wads_register_host', methods=['GET', 'POST'])
@wads_enable
@require_wapt_db
def wads_register_host():
    if request.method == 'POST':
        req = ujson.loads(request.data.decode('utf-8'))
    else:
        req = dict(request.args)

    if app.conf['login_on_wads'] and (not skip_login_wads(request.headers.get('X-Forwarded-For', None))) :
        auth_result = check_auth(session=session, request=request, methods=['admin', 'ldap', 'passwd','token','session'])
        if not auth_result:
            time.sleep(3)
            return authenticate()
        if not has_any_right(get_user_acls(username=auth_result['user']),['admin','wads_host_deploy','wads_admin']):
            return forbidden()

    mac_address = req.get('mac_address', '').upper()
    hostname = req.get('hostname', '')
    wads = req.get('wads', False)

    uuid = req.get('uuid', '')
    manufacturer = req.get('manufacturer', '')
    product = req.get('product', '')
    serial_number  = req.get('serial_number', '')
    asset = req.get('asset', '')
    diskpart_list_disk = req.get('diskpart_list_disk', '')
    diskpart_list_vol = req.get('diskpart_list_vol', '')
    ip_address = req.get('ip_address', '')

    if not wads:
        is_non_alpha = False
        if hostname[0] == '-' or hostname[-1] == '-':
            is_non_alpha = True
        for c in hostname:
            if c == '-':
                continue
            if not c.isalnum():
                is_non_alpha = True
                break
        if len(hostname) > 15 or hostname == '' or is_non_alpha:
            return r"""#!ipxe
clear cmdline
echo Enter valid hostname 15 letters or less for this macaddress
read cmdline
echo Try to register hostname for this macaddress in waptserver
chain %sapi/v3/wads_register_host?mac_address=${net0/mac}&hostname=${cmdline}&uuid=${uuid}&manufacturer=${manufacturer}&product=${product}&serial_number=${serial}&asset=${asset}""" % get_forwarded_url(request)
    # TODO DENY IF MAC OR HOSTNAME ALREADY EXIST

    with WaptDB():
        host = HostOSDeploy.select().where(HostOSDeploy.mac_addresses.contains(mac_address)).first()
        if host and host.mac_addresses:
            return "There is already a host with this mac address. Please check the WAPT console again."

        host_data = dict(
            mac_addresses=mac_address,
            hostname=hostname,
            uuid=uuid,
            manufacturer=manufacturer,
            product=product,
            serial_number=serial_number,
            waiting_to_deploy=False,
            status='registered',
            asset=asset,
            diskpart_list_disk=diskpart_list_disk,
            diskpart_list_vol=diskpart_list_vol,
            ip_address=ip_address
        )

        HostOSDeploy.insert(**host_data).execute()

    logger.info('set waiting to deploy status to false for %s' % (mac_address))

    return "#!ipxe\nreboot"

@wads.route('/api/v3/set_host_wads_status', methods=['POST'])
@wads_enable
@require_wapt_db
def set_host_wads_status():
    status = ujson.loads(request.data)

    mac_address = status["mac_address"].upper()
    host_status = status["host_status"]

    change_host_wads_status(mac_address, host_status)

    return make_response(success=True)

@wads.route('/api/v3/get_wads_exe/<ip>/wads.exe')
def get_wads_exe(ip=''):
    url = get_secondary_repo(ip,get_forwarded_url(request),repo_name='wads')
    return redirect(url+'wads/')

@wads.route('/api/v3/get_wads_exe/<ip>/wads32.exe')
def get_wads_exe32(ip=''):
    url = get_secondary_repo(ip,get_forwarded_url(request),repo_name='wads')
    return redirect(url+'wads/wads32.exe')

@wads.route('/api/v3/get_wads_exe/<ip>/wads64.exe')
def get_wads_exe64(ip=''):
    url = get_secondary_repo(ip,get_forwarded_url(request),repo_name='wads')
    return redirect(url+'wads/wads64.exe')

@wads.route('/api/v3/get_winpe_version', methods=['GET'])
def get_winpe_version():
    filename = os.path.join(app.conf['wads_folder'], 'winpe_version.json')
    if os.path.isfile(filename):
        with open(filename) as f:
            json_version = f.read()
        return make_response(result={'content' : json_version})
    else:
        return make_response(result={'content' : ''}, success=False)
