#!/usr/bin/env python3
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2023  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##

import os

try:
    wapt_root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..'))
except:
    wapt_root_dir = 'c:/tranquilit/wapt'

import shutil
import datetime
import email.utils
import random
import stat
import time
import uuid
import urllib.parse
import requests
import traceback

from waptutils import wget, datetime2isodate, httpdatetime2isodate, httpdatetime2time, fileutcmtime, ensure_unicode, sha1_for_file

from waptserver.config import get_http_proxies

from waptserver.utils import logger

from waptserver.model import WaptDB, WsusDownloadTasks, Hosts, HostWsus, WsusUpdates

from waptserver.wsus_common import check_cab, sha1_from_microsoft_filename, authorized_url_wsus

from waptserver.repositories import update_file_tree_of_files

from waptserver.tasks import huey, conf, crontab
from waptserver.common import licences_list # used as global below. don't delete

from waptserver.tasks_common import conf as tasks_conf

waptwua_folder = tasks_conf['waptwua_folder']


@huey.task(retries=3, retry_delay=60, context=True)
def download_wsusscan(force=False, task=None):
    """Launch a task to update current wsus offline cab file
        download in a temporary well known file
        abort if the temporary file is present (means another download is in progress

    """
    global licences_list
    if not tasks_conf.get('waptwua_enable',True):
        return 'NOT ENABLED'

    if licences_list.is_enterprise():
        with WaptDB():
            in_progress = WsusDownloadTasks.select(WsusDownloadTasks.id).where(
                (WsusDownloadTasks.kind == 'WSUSSCAN') &
                (~WsusDownloadTasks.status.in_(['OK', 'ERROR']))
            )

            if list(in_progress):
                logger.warning('Wsusscan, download already in progress, aborting.')
                return 'SKIPPED'

            cab_url = 'http://download.windowsupdate.com/microsoftupdate/v6/wsusscan/wsusscn2.cab'
            wsus_filename = os.path.join(waptwua_folder, 'wsusscn2.cab')
            tmp_filename = os.path.join(waptwua_folder, 'wsusscn2.cab.tmp_%s' % uuid.uuid1())

            wsusscan2_stats = WsusDownloadTasks(task_id=task.id, skipped=False)
            wsusscan2_stats.kind = 'WSUSSCAN'
            wsusscan2_stats.status = 'INIT'
            wsusscan2_stats.forced = force
            wsusscan2_stats.started_on = datetime.datetime.utcnow()
            wsusscan2_stats.url = cab_url
            wsusscan2_stats.local_filename = tmp_filename
            wsusscan2_stats.save()  # be sure to get an id to not duplicate inserts

            http_proxy = get_http_proxies(conf)

            # should we remove the tmp filename when we get an error?
            cleanup_on_error = False

            if os.path.isfile(tmp_filename):
                os.unlink(tmp_filename)

            try:
                reply = requests.head(cab_url, proxies=http_proxy)
                last_modified = reply.headers['last-modified']
                wsusscan2_stats.file_date = httpdatetime2isodate(last_modified)

                new_cab_timestamp = httpdatetime2time(last_modified)
                if os.path.isfile(wsus_filename):
                    current_cab_timestamp = fileutcmtime(wsus_filename)
                else:
                    current_cab_timestamp = 0.0
                logger.info('download_wsuscan: current cab timestamp: %f, New cab timestamp: %f' % (current_cab_timestamp, new_cab_timestamp))

                if not os.path.isfile(wsus_filename) or (new_cab_timestamp > current_cab_timestamp) or force:
                    logger.info('Downloading because of: file not found == %s, timestamps == %s, force == %s',
                                str(os.path.isfile(wsus_filename) == False), str(new_cab_timestamp > current_cab_timestamp), force)

                    wsusscan2_stats.status = 'DOWNLOADING'
                    wsusscan2_stats.target_size = None
                    if 'content-length' in reply.headers:
                        wsusscan2_stats.target_size = int(reply.headers['content-length'])
                    wsusscan2_stats.file_size = None
                    wsusscan2_stats.save()
                    start_time = time.time()

                    def download_wsusscan_callback(downloaded_bytes, total_bytes, speed, url):
                        wsusscan2_stats.status = 'DOWNLOADING'
                        wsusscan2_stats.file_size = downloaded_bytes
                        wsusscan2_stats.target_size = total_bytes
                        if not wsusscan2_stats.started_on:
                            wsusscan2_stats.started_on = datetime.datetime.utcnow()

                        if downloaded_bytes == 0:
                            wsusscan2_stats.msg = "Downloading %s (%.1f Mb)" % (url, int(total_bytes)/1024/1024)
                        elif downloaded_bytes >= total_bytes:
                            wsusscan2_stats.finished_on = datetime.datetime.utcnow()
                            wsusscan2_stats.msg = "  -> download finished (%.0f Kb/s)" % (total_bytes / (1024.0*(time.time()+.001-start_time)))
                            wsusscan2_stats.status = 'FINISHED'
                        else:
                            wsusscan2_stats.msg = '%i / %i (%.0f%%) (%.0f KB/s)\r' % (downloaded_bytes, total_bytes, 100.0*downloaded_bytes/total_bytes, speed)
                        wsusscan2_stats.save()

                    cleanup_on_error = True
                    wget(cab_url, tmp_filename,
                         printhook=download_wsusscan_callback,
                         resume=True,
                         proxies=http_proxy
                         )

                    file_stats = os.stat(tmp_filename)
                    wsusscan2_stats.file_date = httpdatetime2isodate(email.utils.formatdate(file_stats[stat.ST_MTIME], usegmt=True))
                    wsusscan2_stats.file_size = wsusscan2_stats.target_size = file_stats[stat.ST_SIZE]
                    wsusscan2_stats.status = 'CHECKING'
                    wsusscan2_stats.save()

                    cablist = check_cab(tmp_filename).splitlines()
                    wsusscan2_stats.cablist = cablist

                    if os.path.isfile(wsus_filename):
                        os.unlink(wsus_filename)
                    shutil.move(tmp_filename, wsus_filename)

                    wsusscan2_stats.local_filename = wsus_filename
                    wsusscan2_stats.status = 'OK'
                    wsusscan2_stats.finished_on = datetime2isodate()
                    wsusscan2_stats.save()

                    sumwsus = sha1_for_file(wsus_filename)
                    with open('%s.sha1' % wsus_filename, "w") as f:
                        f.write(sumwsus)

                    if Hosts.select().where(Hosts.wapt_status.contains({'is_remote_repo':True})).count():
                        update_file_tree_of_files(username='TASK_DOWNLOAD_WSUS_SCAN')

                    try:
                        hosts_to_update = Hosts.select(Hosts.uuid, Hosts.waptwua_status).where((~(Hosts.waptwua_status.contains({'status':'NEED-SCAN'}))) & (Hosts.platform == 'Windows'))
                        for the_host in hosts_to_update:
                            if not the_host.waptwua_status is None:
                                if the_host.waptwua_status['wsusscn2cab_date'] != wsusscan2_stats.file_date:
                                    the_host.waptwua_status['status'] = 'NEED-SCAN'
                                    the_host.waptwua_status['wsusscn2cab_date'] = wsusscan2_stats.file_date
                                    Hosts.update(waptwua_status=the_host.waptwua_status).where(Hosts.uuid == the_host.uuid).execute()
                    except Exception as e:
                        logger.debug('Unable to update wsusscn2cab.cab status on hosts: %s' % e)

                    return 'OK'

                else:
                    wsusscan2_stats.finished_on = datetime.datetime.utcnow()
                    wsusscan2_stats.msg = 'Skipped, already up to date'
                    wsusscan2_stats.status = 'OK'
                    wsusscan2_stats.skipped = True
                    wsusscan2_stats.local_filename = wsus_filename
                    file_stats = os.stat(wsusscan2_stats.local_filename)
                    wsusscan2_stats.file_size = file_stats[stat.ST_SIZE]
                    #wsusscan2_stats.finished_on = datetime.datetime.utcnow()
                    wsusscan2_stats.save()

                    if not os.path.isfile('%s.sha1' % wsus_filename):
                        sumwsus = sha1_for_file(wsus_filename)
                        with open('%s.sha1' % wsus_filename, "w") as f:
                            f.write(sumwsus)

                    return 'SKIPPED'

            except Exception as e:
                wsusscan2_stats.status = 'ERROR'
                wsusscan2_stats.error = ensure_unicode(e)
                if not wsusscan2_stats.started_on:
                    wsusscan2_stats.finished_on = datetime.datetime.utcnow()
                wsusscan2_stats.save()

                if cleanup_on_error and os.path.isfile(tmp_filename):
                    os.unlink(tmp_filename)

                logger.error("Error in download_wsusscan: %s", str(e))
                logger.error('Trace:\n%s', traceback.format_exc())
                return 'ERROR'
    else:
        return 'NOT ENTERPRISE'


def get_update_local_path(url):
    """Given a URK, return the local path of file in waptwua folder"""
    fileparts = urllib.parse.urlparse(url).path.split('/')
    target = os.path.join(waptwua_folder, fileparts[-1])
    return target

# Between 2h00 and 2h59


@huey.periodic_task(crontab(hour='3', minute=str(30 + random.randrange(-30, +30))))
def download_wsusscan_crontab():
    global licences_list
    if not tasks_conf.get('waptwua_enable',True):
        return 'DISABLED'
    if licences_list.is_enterprise():
        if tasks_conf['download_wsusscn2']:
            return download_wsusscan()


@huey.periodic_task(crontab(hour='3', minute=str(30 + random.randrange(-30, +30))))
def cleanup_delete_non_assigned_kb():
    global licences_list
    if not tasks_conf.get('waptwua_enable',True):
        return 'NOT ENABLED'
    if licences_list.is_enterprise():
        if tasks_conf['cleanup_kbs']:
            delete_non_assigned_kb('True', 'True')

@huey.task(retries=3, retry_delay=60, context=True)
def download_windows_update_task(url, force=False, task=None):
    """Download a payload for windows updates given its URL

    The URL document part can contains the sha1 of the file.
    In this case, it is checked at the end of the download.

    Args:
        url
        force

    """
    global licences_list
    if not tasks_conf.get('waptwua_enable',True):
        return 'DISABLED'
    if licences_list.is_enterprise():
        with WaptDB():
            target = get_update_local_path(url)
            logger.info(r'Downloading %s to %s' % (url, target))

            # check sha1 sum if possible... return None if no sha1 in filename
            expected_sha1 = sha1_from_microsoft_filename(target)

            http_proxy = get_http_proxies(conf)

            download_task = WsusDownloadTasks(task_id=task.id, skipped=False)
            download_task.kind = 'WINUPDATE'
            download_task.url = url
            download_task.status = 'INIT'
            download_task.local_filename = target
            download_task.started_on = datetime.datetime.utcnow()
            download_task.save()

            start_time = time.time()

            def progress_hook(received, total, speed, url):
                download_task.status = 'DOWNLOADING'
                download_task.file_size = received
                download_task.target_size = total
                if not download_task.started_on:
                    download_task.started_on = datetime.datetime.utcnow()

                if received == 0:
                    download_task.msg = "Downloading %s (%.1f Mb)" % (url, int(total)/1024/1024)
                elif received >= total:
                    download_task.finished_on = datetime.datetime.utcnow()
                    download_task.msg = "  -> download finished (%.0f Kb/s)" % (total / (1024.0*(time.time()+.001-start_time)))
                    download_task.status = 'FINISHED'
                else:
                    download_task.msg = '%i / %i (%.0f%%) (%.0f KB/s)\r' % (received, total, 100.0*received/total, speed)
                download_task.save()
            try:
                wget(url, target, resume=not force, sha1=expected_sha1, proxies=http_proxy, printhook=progress_hook)

                downloaded = True
                download_task.local_filename = target
                download_task.status = 'OK'
                download_task.finished_on = datetime.datetime.utcnow()
                download_task.save()
            except Exception as e:
                download_task.status = 'ERROR'
                download_task.error = ("%s" % e)[:1200]
                download_task.finished_on = datetime.datetime.utcnow()
                download_task.save()

            target_updates = WsusUpdates.filter(WsusUpdates.download_urls.contains(url))
            for update in target_updates:
                for update_url in update.download_urls:
                    if not os.path.isfile(get_update_local_path(update_url)):
                        downloaded = False
                        break

                if downloaded:
                    update.downloaded_on = datetime2isodate()
                    update.save()
                elif not downloaded and update.downloaded_on:
                    update.downloaded_on = None
                    update.save()
            if Hosts.select().where(Hosts.wapt_status.contains({'is_remote_repo':True})).count():
                update_file_tree_of_files(username='TASK_DOWNLOAD_WINDOWS_UPDATE')

            logger.info(r'target file %s correctly downloaded' % target)
            return target
    else:
        return 'NOT ENTERPRISE'


@huey.periodic_task(crontab(minute='*/30'))
def download_missing_cabs():
    """Scans table of Windows updates as reported by hosts.
    If teh Win yupdate shoul dbe installed, but is not yet downloaded, a
    task is created to download the missing files


    """
    # pylint: disable=no-member
    global licences_list
    if not tasks_conf.get('waptwua_enable',True):
        return 'DISABLED'
    if licences_list.is_enterprise():
        with WaptDB():
            updates = (HostWsus.select(WsusUpdates.update_id, WsusUpdates.download_urls)
                       .join(WsusUpdates).distinct()
                       .where((HostWsus.installed == False) & (HostWsus.allowed == True)))
            result = []
            for update in updates:
                for url in update.update_id.download_urls:
                    if authorized_url_wsus(url):
                        local_path = get_update_local_path(url)
                        if not os.path.isfile(local_path) and not url in result:
                            result.append(url)
            for url in result:
                if not authorized_url_wsus(url):
                    logger.info('Url %s not authorized for download' % url)
                    continue
                download_windows_update_task(url)
                logger.info('Queued download of %s cabs' % len(result))
        return result
    else:
        return 'NOT ENTERPRISE'


@huey.lock_task('cleanup-stalled-downloads')
def cleanup_stalled_downloads():
    with WaptDB():
        return (WsusDownloadTasks.update(
            status='ERROR',
            finished_on=datetime.datetime.utcnow(),
            error='Killed, stalled download',
        )
            .where(
                (~WsusDownloadTasks.status.in_(['OK', 'ERROR'])) &
                (~WsusDownloadTasks.task_id.in_([t.task_id for t in huey.pending()]) &
                 (WsusDownloadTasks.started_on < datetime.datetime.utcnow() - datetime.timedelta(hours=4))
                 )
        ).execute()
        )

def delete_non_assigned_kb(need_to_delete, off_the_grid):
    """Uninstall not-needed KBs"""

    with WaptDB():
        data_kbs_to_delete = WsusUpdates.select(WsusUpdates.update_id, WsusUpdates.download_urls).join(HostWsus, 'LEFT JOIN', on=(WsusUpdates.update_id == HostWsus.update_id)).where(
            (HostWsus.update_id.is_null())
        ).tuples()
        data_kbs_to_keep = HostWsus.select(WsusUpdates.update_id, WsusUpdates.download_urls).join(WsusUpdates, 'LEFT JOIN', on=(HostWsus.update_id == WsusUpdates.update_id)).tuples()

        kbs_to_keep = {}
        nb_filename_deleted = 0
        nb_id_cleared = 0
        storage = {}

        for row in data_kbs_to_keep:
            for url in row[1]:
                kbs_to_keep[url.rsplit('/')[-1]] = None

        list_delete_kb_id = []
        for row in data_kbs_to_delete:
            for url in row[1]:
                filename = url.rsplit('/')[-1]
                if not filename in kbs_to_keep:
                    to_delete = os.path.join(tasks_conf['waptwua_folder'], filename)
                    nb_filename_deleted += 1
                    if os.path.isfile(to_delete):
                        if not to_delete in storage and off_the_grid == 'False':
                            storage[to_delete] = os.stat(to_delete).st_size
                        if need_to_delete == 'True':
                            os.unlink(to_delete)
            list_delete_kb_id.append(row[0])
            nb_id_cleared += 1

        data_storage = 0
        if off_the_grid == 'False':
            for files in storage:
                data_storage += storage[files]

        data_storage /= pow(1024, 2)
        if data_storage != 0.0:
            data_storage = '%.3f'%(data_storage)

        if need_to_delete == 'True':
            WsusUpdates.update(downloaded_on=None).where(WsusUpdates.update_id.in_(list_delete_kb_id)).execute()
            if off_the_grid == 'True':
                WsusUpdates.delete().where(WsusUpdates.update_id.in_(list_delete_kb_id)).execute()

        result = len(list_delete_kb_id)
        if result == 0:
            return [result]
        elif off_the_grid == 'True':
            return [nb_id_cleared, data_storage]
        else:
            return [nb_filename_deleted, data_storage]
